#!/usr/bin/env python
# stun_server.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (stun_server.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: stun_server.

BitDust stun_server() Automat

EVENTS:
    * :red:`datagram-received`
    * :red:`start`
    * :red:`stop`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

import sys

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio

from bitdust.main import settings

from bitdust.automats import automat

from bitdust.lib import net_misc
from bitdust.lib import udp

from bitdust.dht import dht_service

#------------------------------------------------------------------------------

_StunServer = None

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _StunServer
    if _StunServer is None:
        # set automat name and starting state here
        _StunServer = StunServer(
            name='stun_server',
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
        )
    if event is not None:
        _StunServer.automat(event, *args, **kwargs)
    return _StunServer


def Destroy():
    """
    Destroy stun_server() automat and remove its instance from memory.
    """
    global _StunServer
    if _StunServer is None:
        return
    _StunServer.destroy()
    del _StunServer
    _StunServer = None


class StunServer(automat.Automat):
    """
    This class implements all the functionality of the ``stun_server()`` state
    machine.
    """

    fast = True

    def init(self):
        self.listen_port = None

    def A(self, event, *args, **kwargs):
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'start':
                self.state = 'LISTEN'
                self.doInit(*args, **kwargs)
        #---LISTEN---
        elif self.state == 'LISTEN':
            if event == 'stop':
                self.state = 'STOPPED'
                self.doStop(*args, **kwargs)
            elif event == 'datagram-received' and self.isSTUN(*args, **kwargs):
                self.doSendYourIPPort(*args, **kwargs)
        #---STOPPED---
        elif self.state == 'STOPPED':
            if event == 'start':
                self.state = 'LISTEN'
                self.doInit(*args, **kwargs)

    def isSTUN(self, *args, **kwargs):
        """
        Condition method.
        """
        try:
            datagram, address = args[0]
            command, payload = datagram
        except:
            return False
        return command == udp.CMD_STUN

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        self.listen_port = args[0]
        if udp.proto(self.listen_port):
            udp.proto(self.listen_port).add_callback(self._datagramReceived)
            lg.info('callback added to listen on UDP port %d' % self.listen_port)
        else:
            lg.err('udp port %s is not opened' % self.listen_port)
        # try:
        #     externalPort = int(bpio.ReadTextFile(settings.ExternalUDPPortFilename()))
        # except:
        #     lg.exc()
        externalPort = self.listen_port
        dht_service.set_node_data('stun_port', externalPort)

    def doStop(self, *args, **kwargs):
        """
        Action method.
        """
        if udp.proto(self.listen_port):
            udp.proto(self.listen_port).remove_callback(self._datagramReceived)
        else:
            lg.err('udp port %s is not opened' % self.listen_port)

    def doSendYourIPPort(self, *args, **kwargs):
        """
        Action method.
        """
        try:
            datagram, address = args[0]
            command, payload = datagram
        except:
            return False
        youripport = net_misc.pack_address((address[0], address[1]))
        udp.send_command(self.listen_port, udp.CMD_MYIPPORT, youripport, address)
        if _Debug:
            lg.out(_DebugLevel, 'stun_server.doSendYourIPPort [%s] to %s' % (youripport, address))

    def _datagramReceived(self, datagram, address):
        """
        """
        self.automat('datagram-received', (datagram, address))
        return False


#------------------------------------------------------------------------------


def main():
    from twisted.internet import reactor  # @UnresolvedImport
    lg.set_debug_level(24)
    bpio.init()
    settings.init()
    dht_port = settings.getDHTPort()
    if len(sys.argv) > 1:
        dht_port = int(sys.argv[1])
    udp_port = settings.getUDPPort()
    if len(sys.argv) > 2:
        udp_port = int(sys.argv[2])
    dht_service.init(dht_port)
    d = dht_service.connect()
    udp.listen(udp_port)

    def _go(live_nodes):
        A('start', udp_port)

    d.addCallback(_go)
    reactor.run()  # @UndefinedVariable
    settings.shutdown()


#------------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# shared_access_coordinator.py
#
# Copyright (C) 2008 Veselin Penev, http://bitdust.io
#
# This file (shared_access_coordinator.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: shared_access_coordinator
.. role:: red

BitDust shared_access_coordinator() Automat

EVENTS:
    * :red:`all-suppliers-connected`
    * :red:`all-suppliers-disconnected`
    * :red:`dht-lookup-ok`
    * :red:`fail`
    * :red:`index-failed`
    * :red:`index-missing`
    * :red:`index-received`
    * :red:`index-sent`
    * :red:`index-up-to-date`
    * :red:`key-not-registered`
    * :red:`key-sent`
    * :red:`list-files-failed`
    * :red:`list-files-received`
    * :red:`list-files-verified`
    * :red:`new-private-key-registered`
    * :red:`restart`
    * :red:`shutdown`
    * :red:`supplier-connected`
    * :red:`supplier-failed`
    * :red:`supplier-file-modified`
    * :red:`timer-1min`
    * :red:`timer-30sec`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import time
import base64

#------------------------------------------------------------------------------

from twisted.internet import reactor  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.system import bpio

from bitdust.lib import strng
from bitdust.lib import serialization
from bitdust.lib import packetid

from bitdust.main import events
from bitdust.main import settings
from bitdust.main import listeners

from bitdust.dht import dht_relations

from bitdust.p2p import commands
from bitdust.p2p import p2p_service
from bitdust.p2p import online_status

from bitdust.contacts import identitycache
from bitdust.contacts import contactsdb

from bitdust.crypt import my_keys
from bitdust.crypt import key
from bitdust.crypt import signed
from bitdust.crypt import encrypted

from bitdust.access import key_ring

from bitdust.storage import backup_control
from bitdust.storage import backup_matrix
from bitdust.storage import backup_fs

from bitdust.customer import supplier_connector

from bitdust.userid import global_id
from bitdust.userid import id_url
from bitdust.userid import my_id

#------------------------------------------------------------------------------

_ActiveShares = {}
_ActiveSharesByIDURL = {}

#------------------------------------------------------------------------------


def register_share(A):
    global _ActiveShares
    global _ActiveSharesByIDURL
    if A.key_id in _ActiveShares:
        raise Exception('share already exist')
    if id_url.is_not_in(A.customer_idurl, _ActiveSharesByIDURL):
        _ActiveSharesByIDURL[A.customer_idurl] = []
    _ActiveSharesByIDURL[A.customer_idurl].append(A)
    _ActiveShares[A.key_id] = A


def unregister_share(A):
    global _ActiveShares
    global _ActiveSharesByIDURL
    _ActiveShares.pop(A.key_id, None)
    if id_url.is_not_in(A.customer_idurl, _ActiveSharesByIDURL):
        lg.warn('given customer idurl not found in active shares list')
    else:
        _ActiveSharesByIDURL[A.customer_idurl] = []


#------------------------------------------------------------------------------


def list_active_shares():
    global _ActiveShares
    return list(_ActiveShares.keys())


def get_active_share(key_id):
    global _ActiveShares
    if key_id not in _ActiveShares:
        return None
    return _ActiveShares[key_id]


def find_active_shares(customer_idurl):
    global _ActiveSharesByIDURL
    result = []
    for automat_index in _ActiveSharesByIDURL.values():
        A = automat.by_index(automat_index)
        if not A:
            continue
        if A.customer_idurl == customer_idurl:
            result.append(A)
    return result


#-----------------------------------------------------------------------------


def populate_shares():
    global _ActiveShares
    for share_instance in _ActiveShares.values():
        listeners.push_snapshot('shared_location', snap_id=share_instance.key_id, data=share_instance.to_json())


#------------------------------------------------------------------------------


def open_known_shares():
    known_offline_shares = []
    for key_id in my_keys.known_keys():
        if not key_id.startswith('share_'):
            continue
        if not my_keys.is_active(key_id):
            continue
        active_share = get_active_share(key_id)
        if active_share:
            continue
        known_offline_shares.append(key_id)
    to_be_opened = []
    for customer_idurl in backup_fs.known_customers():
        for key_alias in backup_fs.known_keys_aliases(customer_idurl):
            if not key_alias.startswith('share_'):
                continue
            key_id = my_keys.make_key_id(alias=key_alias, creator_idurl=customer_idurl)
            if not key_id:
                continue
            if my_keys.latest_key_id(key_id) != key_id:
                continue
            if key_id in to_be_opened:
                continue
            if key_id not in known_offline_shares:
                continue
            if not my_keys.is_key_private(key_id):
                continue
            if not my_keys.is_active(key_id):
                continue
            to_be_opened.append(key_id)
    if _Debug:
        lg.args(_DebugLevel, known_offline_shares=known_offline_shares, to_be_opened=to_be_opened)
    populate_shared_files = listeners.is_populate_required('shared_file')
    for key_id in to_be_opened:
        active_share = SharedAccessCoordinator(key_id, log_events=True, publish_events=False)
        active_share.automat('restart')
        if populate_shared_files:
            backup_fs.populate_shared_files(key_id=key_id)
    # if populate_shared_files:
    # listeners.populate_later().remove('shared_file')
    if listeners.is_populate_required('shared_location'):
        # listeners.populate_later().remove('shared_location')
        populate_shares()


#------------------------------------------------------------------------------


def get_deleted_path_ids(customer_idurl, key_alias):
    key_id = global_id.MakeGlobalID(idurl=customer_idurl, key_alias=key_alias)
    if not my_keys.is_key_registered(key_id):
        return []
    if not my_keys.is_active(key_id):
        return []
    active_share = get_active_share(key_id)
    if not active_share:
        return []
    if _Debug:
        lg.args(_DebugLevel, k=key_id, ret=active_share.files_to_be_deleted)
    return active_share.files_to_be_deleted


#------------------------------------------------------------------------------


def on_file_deleted(customer_idurl, key_alias, path_id):
    key_id = global_id.MakeGlobalID(idurl=customer_idurl, key_alias=key_alias)
    if _Debug:
        lg.args(_DebugLevel, k=key_id, path=path_id)
    if not my_keys.is_key_registered(key_id):
        return
    if not my_keys.is_active(key_id):
        return
    active_share = get_active_share(key_id)
    if not active_share:
        lg.warn('index file was updated and key is active, but share %s is not known' % key_id)
        return
    active_share.files_to_be_deleted.append(path_id)


def on_index_file_updated(customer_idurl, key_alias):
    key_id = global_id.MakeGlobalID(idurl=customer_idurl, key_alias=key_alias)
    if _Debug:
        lg.args(_DebugLevel, k=key_id)
    if not my_keys.is_key_registered(key_id):
        return
    if not my_keys.is_active(key_id):
        return
    active_share = get_active_share(key_id)
    if not active_share:
        lg.warn('index file was updated and key is active, but share %s is not known' % key_id)
        return
    if active_share.state == 'DISCONNECTED':
        active_share.automat('restart')
        return
    if active_share.state != 'CONNECTED':
        active_share.to_be_restarted = True
        return
    active_share.automat('restart')


def on_supplier_file_modified(evt):
    if evt.data['key_alias'] == 'master':
        return
    key_id = global_id.MakeGlobalID(idurl=evt.data['customer_idurl'], key_alias=evt.data['key_alias'])
    if _Debug:
        lg.args(_DebugLevel, e=evt, d=evt.data, k=key_id)
    if not my_keys.is_key_registered(key_id):
        return
    if not my_keys.is_active(key_id):
        return
    active_share = get_active_share(key_id)
    if not active_share:
        lg.warn('supplier file was modified and key is active, but share %s is not known' % key_id)
        return
    if active_share.state == 'DISCONNECTED':
        active_share.automat('restart')
        return
    # if active_share.state != 'CONNECTED':
    #     active_share.to_be_restarted = True
    active_share.automat('supplier-file-modified', supplier_idurl=evt.data['supplier_idurl'], remote_path=evt.data['remote_path'])


def on_key_registered(evt):
    if not evt.data['key_id'].startswith('share_'):
        return
    active_share = get_active_share(evt.data['key_id'])
    if _Debug:
        lg.args(_DebugLevel, e=evt, active_share=active_share)
    if active_share:
        active_share.automat('new-private-key-registered')
        return

    def _run_coordinator():
        new_share = SharedAccessCoordinator(
            key_id=evt.data['key_id'],
            log_events=True,
            publish_events=False,
        )
        new_share.add_connected_callback('key_registered' + strng.to_text(time.time()), lambda _id, _result: on_share_first_connected(evt.data['key_id'], _id, _result))
        new_share.automat('new-private-key-registered')

    glob_id = global_id.NormalizeGlobalID(evt.data['key_id'])
    if id_url.is_cached(glob_id['idurl']):
        _run_coordinator()
    else:
        d = identitycache.immediatelyCaching(glob_id['idurl'])
        d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='shared_access_coordinator.on_key_registered')
        d.addCallback(lambda *args: _run_coordinator())


def on_key_erased(evt):
    if not evt.data['key_id'].startswith('share_'):
        return
    from bitdust.access import shared_access_coordinator
    active_share = shared_access_coordinator.get_active_share(evt.data['key_id'])
    if _Debug:
        lg.args(_DebugLevel, e=evt, active_share=active_share)
    if active_share:
        active_share.automat('shutdown')


def on_share_connected(evt):
    if _Debug:
        lg.args(_DebugLevel, e=evt)


def on_share_first_connected(key_id, callback_id, result):
    if not result:
        return
    active_share = get_active_share(key_id)
    if _Debug:
        lg.args(_DebugLevel, key_id=key_id, result=result, active_share=active_share)
    if active_share:
        active_share.remove_connected_callback(callback_id)
        backup_fs.populate_shared_files(key_id=key_id)


def on_supplier_modified(evt):
    if _Debug:
        lg.args(_DebugLevel, e=evt)
    if evt.data['new_idurl']:
        my_keys_to_be_republished = []
        for key_id in my_keys.known_keys():
            if not key_id.startswith('share_'):
                continue
            _glob_id = global_id.NormalizeGlobalID(key_id)
            if _glob_id['idurl'].to_bin() == my_id.getIDURL().to_bin():
                # only send public keys of my own shares
                my_keys_to_be_republished.append(key_id)
        for key_id in my_keys_to_be_republished:
            d = key_ring.transfer_key(key_id, trusted_idurl=evt.data['new_idurl'], include_private=False, include_signature=False)
            d.addErrback(lambda *a: lg.err('transfer key failed: %s' % str(*a)))


def on_my_list_files_refreshed(evt):
    if _Debug:
        lg.args(_DebugLevel, e=evt)
    for key_id in list_active_shares():
        cur_share = get_active_share(key_id)
        if not cur_share:
            continue
        if cur_share.state == 'DISCONNECTED':
            cur_share.automat('restart')
            continue
        if cur_share.state != 'CONNECTED':
            continue
        if not cur_share.connected_last_time:
            cur_share.automat('restart')
            continue
        if time.time() - cur_share.connected_last_time < 60:
            continue
        cur_share.automat('restart')


def on_list_files_verified(newpacket, list_files_info):
    incoming_key_id = list_files_info['key_id']
    # incoming_key_alias = list_files_info['key_alias']
    active_share = get_active_share(incoming_key_id)
    if not active_share:
        lg.warn('active share was not found for incoming key %r' % incoming_key_id)
        return False
    if active_share.state == 'DISCONNECTED':
        lg.warn('active share is currently disconnect for incoming key %r, restarting' % incoming_key_id)
        active_share.automat('restart')
        return False
    try:
        block = encrypted.Unserialize(
            newpacket.Payload,
            decrypt_key=incoming_key_id,
        )
    except:
        lg.exc(newpacket.Payload)
        return False
    if block is None:
        lg.warn('failed reading data from %s' % newpacket.RemoteID)
        return False
    try:
        raw_files = block.Data()
    except:
        lg.exc()
        return False
    # otherwise this must be an external supplier sending us a files he stores for trusted customer
    external_supplier_idurl = block.CreatorID
    #     supplier_index_file_revision = active_share.received_index_file_revision.get(external_supplier_idurl)
    #     if supplier_index_file_revision:
    #         _rev = backup_fs.revision(customer_idurl=active_share.customer_idurl, key_alias=incoming_key_alias)
    #         if supplier_index_file_revision <= _rev:
    #             lg.warn('shared location index file is not in sync, local revision is %r but revision by supplier is %r' % (_rev, supplier_index_file_revision))
    #             return False
    try:
        supplier_raw_list_files = backup_control.UnpackListFiles(raw_files, settings.ListFilesFormat())
    except:
        lg.exc()
        return False
    # need to detect supplier position from the list of packets
    # and place that supplier on the correct position in contactsdb
    supplier_pos = backup_matrix.DetectSupplierPosition(supplier_raw_list_files)
    known_supplier_pos = contactsdb.supplier_position(external_supplier_idurl, active_share.customer_idurl)
    if known_supplier_pos < 0:
        lg.warn('received %r from an unknown node %r which is not a supplier of %r' % (newpacket, external_supplier_idurl, active_share.customer_idurl))
        return False
    if _Debug:
        lg.args(_DebugLevel, sz=len(supplier_raw_list_files), s=external_supplier_idurl, pos=supplier_pos, known_pos=known_supplier_pos, pid=newpacket.PacketID)
    if supplier_pos >= 0:
        if known_supplier_pos != supplier_pos:
            lg.err('known external supplier %r position %d is not matching with received list files position %d for customer %s' % (external_supplier_idurl, known_supplier_pos, supplier_pos, active_share.customer_idurl))
            return False
    else:
        lg.warn('not possible to detect external supplier position for customer %s from received list files, known position is %s' % (active_share.customer_idurl, known_supplier_pos))
        supplier_pos = known_supplier_pos
    active_share.automat(
        'list-files-verified',
        newpacket=newpacket,
        supplier_pos=supplier_pos,
        supplier_idurl=external_supplier_idurl,
        payload=supplier_raw_list_files,
    )
    # finally sending Ack() packet back
    p2p_service.SendAck(newpacket)
    return True


#------------------------------------------------------------------------------


class SharedAccessCoordinator(automat.Automat):

    """
    This class implements all the functionality of the ``shared_access_coordinator()`` state machine.
    """
    fast = False

    timers = {
        'timer-1min': (60, ['DHT_LOOKUP']),
        'timer-30sec': (30.0, ['SUPPLIERS?']),
    }

    def __init__(self, key_id, debug_level=_DebugLevel, log_events=_Debug, log_transitions=_Debug, publish_events=False, **kwargs):
        """
        Create shared_access_coordinator() state machine.
        Use this method if you need to call Automat.__init__() in a special way.
        """
        self.key_id = key_id
        self.glob_id = global_id.NormalizeGlobalID(self.key_id)
        self.key_alias = self.glob_id['key_alias']
        self.customer_idurl = self.glob_id['idurl']
        self.known_suppliers_list = []
        self.known_ecc_map = None
        self.dht_lookup_use_cache = True
        self.received_index_file_revision = {}
        self.last_time_in_sync = -1
        self.suppliers_in_progress = []
        self.suppliers_succeed = []
        self.to_be_restarted = False
        self.files_to_be_deleted = []
        self.connected_last_time = None
        super(SharedAccessCoordinator, self).__init__(
            name='%s$%s' % (self.key_alias, self.glob_id['customer']),
            state='AT_STARTUP',
            debug_level=debug_level,
            log_events=log_events,
            log_transitions=log_transitions,
            publish_events=publish_events,
            **kwargs,
        )

    def to_json(self):
        j = super().to_json()
        j.update(
            {
                'active': my_keys.is_active(self.key_id),
                'key_id': self.key_id,
                'alias': self.key_alias,
                'label': my_keys.get_label(self.key_id) or '',
                'creator': self.customer_idurl.to_id(),
                'suppliers': [id_url.idurl_to_id(s) for s in self.known_suppliers_list],
                'ecc_map': self.known_ecc_map,
                'revision': backup_fs.revision(self.customer_idurl, self.key_alias),
            }
        )
        return j

    def add_connected_callback(self, callback_id, callback_method):
        self.connected_callbacks[callback_id] = callback_method

    def remove_connected_callback(self, callback_id):
        self.connected_callbacks.pop(callback_id, None)

    def init(self):
        """
        Method to initialize additional variables and flags
        at creation phase of shared_access_coordinator() machine.
        """
        self.result_defer = None
        self.connected_callbacks = {}

    def register(self):
        automat_index = automat.Automat.register(self)
        register_share(self)
        return automat_index

    def unregister(self):
        unregister_share(self)
        return automat.Automat.unregister(self)

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        """
        Method to catch the moment when `shared_access_coordinator()` state were changed.
        """
        if _Debug:
            lg.out(_DebugLevel, '%s : [%s]->[%s]' % (self.name, oldstate, newstate))
        if newstate == 'CONNECTED':
            lg.info('share connected : %s' % self.key_id)
            self.connected_last_time = time.time()
            listeners.push_snapshot('shared_location', snap_id=self.key_id, data=self.to_json())
            self.files_to_be_deleted = []
            if self.to_be_restarted:
                self.to_be_restarted = False
                reactor.callLater(1, self.automat, 'restart')  # @UndefinedVariable
        elif newstate == 'DISCONNECTED' and oldstate != 'AT_STARTUP':
            lg.info('share disconnected : %s' % self.key_id)
            self.connected_last_time = None
            listeners.push_snapshot('shared_location', snap_id=self.key_id, data=self.to_json())
            if self.to_be_restarted:
                self.to_be_restarted = False
                reactor.callLater(1, self.automat, 'restart')  # @UndefinedVariable
        elif newstate in ['DHT_LOOKUP', 'SUPPLIERS?', 'CLOSED'] and oldstate != 'AT_STARTUP':
            listeners.push_snapshot('shared_location', snap_id=self.key_id, data=self.to_json())

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python <http://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'restart' or event == 'new-private-key-registered':
                self.state = 'DHT_LOOKUP'
                self.doInit(*args, **kwargs)
                self.doDHTLookupSuppliers(*args, **kwargs)
        #---DHT_LOOKUP---
        elif self.state == 'DHT_LOOKUP':
            if event == 'dht-lookup-ok':
                self.state = 'SUPPLIERS?'
                self.doConnectCustomerSuppliers(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'fail' or event == 'timer-1min':
                self.state = 'DISCONNECTED'
                self.doReportDisconnected(*args, **kwargs)
        #---SUPPLIERS?---
        elif self.state == 'SUPPLIERS?':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'timer-30sec' or event == 'all-suppliers-disconnected':
                self.state = 'DISCONNECTED'
                self.doReportDisconnected(*args, **kwargs)
            elif event == 'list-files-received':
                self.doSupplierSendIndexFile(*args, **kwargs)
            elif event == 'key-not-registered':
                self.doSupplierTransferPubKey(*args, **kwargs)
            elif event == 'supplier-connected' or event == 'key-sent':
                self.doSupplierRequestIndexFile(*args, **kwargs)
            elif event == 'all-suppliers-connected':
                self.state = 'CONNECTED'
                self.doReportConnected(*args, **kwargs)
            elif event == 'index-sent' or event == 'index-up-to-date' or event == 'index-failed' or event == 'list-files-failed' or event == 'supplier-failed':
                self.doRemember(event, *args, **kwargs)
                self.doCheckAllConnected(*args, **kwargs)
            elif event == 'list-files-verified':
                self.doSupplierProcessListFiles(*args, **kwargs)
            elif event == 'supplier-file-modified' or event == 'index-received' or event == 'index-missing':
                self.doSupplierRequestListFiles(event, *args, **kwargs)
        #---DISCONNECTED---
        elif self.state == 'DISCONNECTED':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'restart':
                self.state = 'DHT_LOOKUP'
                self.doDHTLookupSuppliers(*args, **kwargs)
        #---CONNECTED---
        elif self.state == 'CONNECTED':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'supplier-failed' or event == 'restart':
                self.state = 'DHT_LOOKUP'
                self.doDHTLookupSuppliers(*args, **kwargs)
            elif event == 'supplier-file-modified' or event == 'index-received' or event == 'index-missing':
                self.doSupplierRequestListFiles(event, *args, **kwargs)
            elif event == 'list-files-verified':
                self.doSupplierProcessListFiles(*args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        return None

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        # TODO : put in a separate state in the state machine
        self.result_defer = kwargs.get('result_defer', None)
        identitycache.immediatelyCaching(self.customer_idurl, ignore_errors=True)

    def doDHTLookupSuppliers(self, *args, **kwargs):
        """
        Action method.
        """
        d = dht_relations.read_customer_suppliers(self.customer_idurl, use_cache=self.dht_lookup_use_cache)
        # TODO: add more validations of dht_result
        d.addCallback(self._on_read_customer_suppliers)
        d.addErrback(lambda err: self.automat('fail', err))

    def doConnectCustomerSuppliers(self, *args, **kwargs):
        """
        Action method.
        """
        try:
            self.known_suppliers_list = [s for s in args[0]['suppliers'] if s]
        except:
            lg.exc()
            self.automat('all-suppliers-disconnected')
            return
        self.known_ecc_map = args[0].get('ecc_map')
        self.suppliers_in_progress.clear()
        self.suppliers_succeed.clear()
        for supplier_idurl in self.known_suppliers_list:
            self.suppliers_in_progress.append(supplier_idurl)
            if id_url.is_cached(supplier_idurl):
                self._do_connect_with_supplier(supplier_idurl)
            else:
                d = identitycache.immediatelyCaching(supplier_idurl)
                d.addCallback(lambda *a: self._do_connect_with_supplier(supplier_idurl))
                d.addErrback(self._on_supplier_failed, supplier_idurl=kwargs['supplier_idurl'], reason='failed caching supplier identity')
        if _Debug:
            lg.args(_DebugLevel, known_ecc_map=self.known_ecc_map, known_suppliers_list=self.known_suppliers_list)

    def doSupplierRequestIndexFile(self, *args, **kwargs):
        """
        Action method.
        """
        self._do_retrieve_index_file(kwargs['supplier_idurl'])

    def doSupplierRequestListFiles(self, event, *args, **kwargs):
        """
        Action method.
        """
        supplier_idurl = kwargs['supplier_idurl']
        pkt_out = None
        if event == 'supplier-file-modified':
            remote_path = kwargs['remote_path']
            if remote_path == settings.BackupIndexFileName():
                if self.state == 'CONNECTED':
                    self.automat('restart')
                else:
                    self.to_be_restarted = True
            else:
                _, remote_path, _, _ = packetid.SplitVersionFilename(remote_path)
                iter_path = backup_fs.WalkByID(remote_path, iterID=backup_fs.fsID(self.customer_idurl, self.key_alias))
                if not iter_path:
                    lg.warn('did not found modified file %r in the catalog, restarting %r' % (kwargs['remote_path'], self))
                    self.automat('restart')
                else:
                    sc = supplier_connector.by_idurl(
                        supplier_idurl,
                        customer_idurl=self.customer_idurl,
                    )
                    if sc is not None and sc.state == 'CONNECTED':
                        pkt_out = p2p_service.SendListFiles(
                            target_supplier=supplier_idurl,
                            customer_idurl=self.customer_idurl,
                            key_id=self.key_id,
                            timeout=settings.P2PTimeOut(),
                            callbacks={
                                commands.Files(): lambda r, i: self._on_list_files_response(r, i, self.customer_idurl, supplier_idurl, self.key_id),
                                commands.Fail(): lambda r, i: self._on_list_files_failed(r, i, self.customer_idurl, supplier_idurl, self.key_id),
                                None: lambda pkt_out: self._on_list_files_timeout(self.customer_idurl, supplier_idurl, self.key_id),
                            },
                        )
        else:
            pkt_out = p2p_service.SendListFiles(
                target_supplier=supplier_idurl,
                customer_idurl=self.customer_idurl,
                key_id=self.key_id,
                timeout=settings.P2PTimeOut(),
                callbacks={
                    commands.Files(): lambda r, i: self._on_list_files_response(r, i, self.customer_idurl, supplier_idurl, self.key_id),
                    commands.Fail(): lambda r, i: self._on_list_files_failed(r, i, self.customer_idurl, supplier_idurl, self.key_id),
                    None: lambda pkt_out: self._on_list_files_timeout(self.customer_idurl, supplier_idurl, self.key_id),
                },
            )
        if _Debug:
            lg.args(_DebugLevel, e=event, s=supplier_idurl, outgoing=pkt_out)

    def doSupplierTransferPubKey(self, *args, **kwargs):
        """
        Action method.
        """
        d = key_ring.transfer_key(kwargs['key_id'], kwargs['supplier_idurl'], include_private=False, include_signature=False)
        d.addCallback(lambda r: self._on_key_transfer_success(**kwargs))
        d.addErrback(self._on_supplier_failed, supplier_idurl=kwargs['supplier_idurl'], reason='failed sending key %r' % kwargs['key_id'])

    def doSupplierSendIndexFile(self, *args, **kwargs):
        """
        Action method.
        """
        supplier_index_file_revision = self.received_index_file_revision.get(kwargs['supplier_idurl'])
        if supplier_index_file_revision:
            _rev = backup_fs.revision(customer_idurl=self.customer_idurl, key_alias=self.key_alias)
            if _Debug:
                lg.args(_DebugLevel, s=kwargs['supplier_idurl'], supplier_rev=supplier_index_file_revision, my_rev=_rev)
            if supplier_index_file_revision >= _rev:
                self.automat('index-up-to-date', supplier_idurl=kwargs['supplier_idurl'])
                return
        self._do_send_index_file(kwargs['supplier_idurl'])

    def doSupplierProcessListFiles(self, *args, **kwargs):
        """
        Action method.
        """
        is_in_sync = False
        supplier_index_file_revision = self.received_index_file_revision.get(kwargs['supplier_idurl'])
        if supplier_index_file_revision:
            _rev = backup_fs.revision(customer_idurl=self.customer_idurl, key_alias=self.key_alias)
            if supplier_index_file_revision <= _rev:
                is_in_sync = True
        if _Debug:
            lg.args(_DebugLevel, rev=supplier_index_file_revision, is_in_sync=is_in_sync)
        remote_files_changed, backups2remove, paths2remove, missed_backups = backup_matrix.process_raw_list_files(
            supplier_num=kwargs['supplier_pos'],
            list_files_text_body=kwargs['payload'],
            customer_idurl=self.customer_idurl,
            is_in_sync=is_in_sync,
        )
        if remote_files_changed:
            backup_matrix.SaveLatestRawListFiles(
                supplier_idurl=kwargs['supplier_idurl'],
                raw_data=kwargs['payload'],
                customer_idurl=self.customer_idurl,
            )
            if _Debug:
                lg.dbg(_DebugLevel, 'received updated list of files from external supplier %s for customer %s' % (kwargs['supplier_idurl'], self.customer_idurl))
            if len(backups2remove) > 0:
                p2p_service.RequestDeleteListBackups(backups2remove)
                if _Debug:
                    lg.out(_DebugLevel, '    also sent requests to remove %d shared backups' % len(backups2remove))
            if len(paths2remove) > 0:
                p2p_service.RequestDeleteListPaths(paths2remove)
                if _Debug:
                    lg.out(_DebugLevel, '    also sent requests to remove %d shared paths' % len(paths2remove))

    def doRemember(self, event, *args, **kwargs):
        """
        Action method.
        """
        supplier_idurl = kwargs['supplier_idurl']
        if supplier_idurl in self.suppliers_in_progress:
            self.suppliers_in_progress.remove(supplier_idurl)
            if event in ['index-sent', 'index-up-to-date']:
                if supplier_idurl not in self.suppliers_succeed:
                    self.suppliers_succeed.append(supplier_idurl)
        if _Debug:
            lg.args(_DebugLevel, e=event, s=supplier_idurl, progress=len(self.suppliers_in_progress), succeed=len(self.suppliers_succeed))

    def doCheckAllConnected(self, *args, **kwargs):
        """
        Action method.
        """
        critical_suppliers_number = 1
        if self.known_ecc_map:
            from bitdust.raid import eccmap
            critical_suppliers_number = eccmap.GetCorrectableErrors(eccmap.GetEccMapSuppliersNumber(self.known_ecc_map))
        if _Debug:
            lg.args(_DebugLevel, progress=len(self.suppliers_in_progress), succeed=self.suppliers_succeed, critical_suppliers_number=critical_suppliers_number)
        if len(self.suppliers_in_progress) == 0:
            if len(self.suppliers_succeed) >= critical_suppliers_number:
                self.automat('all-suppliers-connected')
            else:
                self.automat('all-suppliers-disconnected')

    def doReportConnected(self, *args, **kwargs):
        """
        Action method.
        """
        self.dht_lookup_use_cache = True
        events.send('share-connected', data=dict(self.to_json()))
        if self.result_defer:
            self.result_defer.callback(True)
        for cb_id in list(self.connected_callbacks.keys()):
            cb = self.connected_callbacks.get(cb_id)
            if cb:
                cb(cb_id, True)
        if _Debug:
            lg.args(_DebugLevel, key_id=self.key_id, ecc_map=self.known_ecc_map)

    def doReportDisconnected(self, *args, **kwargs):
        """
        Action method.
        """
        self.dht_lookup_use_cache = False
        events.send('share-disconnected', data=dict(self.to_json()))
        if self.result_defer:
            self.result_defer.errback(Exception('disconnected'))
        for cb_id in list(self.connected_callbacks.keys()):
            if cb_id in self.connected_callbacks:
                cb = self.connected_callbacks[cb_id]
                cb(cb_id, False)
        if _Debug:
            lg.args(_DebugLevel, key_id=self.key_id, ecc_map=self.known_ecc_map)

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        self.result_defer = None
        self.destroy()

    def _do_connect_with_supplier(self, supplier_idurl):
        if _Debug:
            lg.args(_DebugLevel, supplier_idurl=supplier_idurl, customer_idurl=self.customer_idurl)
        sc = supplier_connector.by_idurl(supplier_idurl, customer_idurl=self.customer_idurl)
        if sc is None:
            sc = supplier_connector.create(
                supplier_idurl=supplier_idurl,
                customer_idurl=self.customer_idurl,
                needed_bytes=0,  # we only want to read the data at the moment - requesting 0 bytes from the supplier
                key_id=self.key_id,
                queue_subscribe=True,
            )
        if sc.state in ['CONNECTED', 'QUEUE?']:
            self.automat('supplier-connected', supplier_idurl=supplier_idurl)
        else:
            sc.set_callback('shared_access_coordinator', self._on_supplier_connector_state_changed)
            sc.automat('connect')

    def _do_retrieve_index_file(self, supplier_idurl):
        packetID = global_id.MakeGlobalID(
            key_id=self.key_id,
            path=settings.BackupIndexFileName(),
        )
        sc = supplier_connector.by_idurl(supplier_idurl, customer_idurl=self.customer_idurl)
        if sc is None or sc.state != 'CONNECTED':
            lg.warn('supplier connector for %r is not found or offline' % supplier_idurl)
            self.automat('supplier-failed', supplier_idurl=supplier_idurl)
            return
        if online_status.isOffline(supplier_idurl):
            lg.warn('supplier %r is offline' % supplier_idurl)
            self.automat('supplier-failed', supplier_idurl=supplier_idurl)
            return
        public_test_sample = key.NewSessionKey(session_key_type=key.SessionKeyType())
        signed_test_sample = my_keys.sign(self.key_id, public_test_sample)
        json_payload = {
            't': base64.b64encode(public_test_sample),
            's': strng.to_text(signed_test_sample),
        }
        raw_payload = serialization.DictToBytes(json_payload, values_to_text=True)
        p2p_service.SendRetreive(
            ownerID=self.customer_idurl,
            creatorID=my_id.getIDURL(),
            packetID=packetID,
            remoteID=supplier_idurl,
            response_timeout=settings.P2PTimeOut(),
            payload=raw_payload,
            callbacks={
                commands.Data(): self._on_index_file_response,
                commands.Fail(): self._on_index_file_fail_received,
                None: lambda pkt_out: self._on_index_file_request_failed(supplier_idurl, self.customer_idurl, packetID),
                'failed': lambda pkt_out, errmsg: self._on_index_file_request_failed(supplier_idurl, self.customer_idurl, packetID),
            },
        )
        if _Debug:
            lg.args(_DebugLevel, pid=packetID, supplier=supplier_idurl)

    def _do_process_index_file(self, wrapped_packet, supplier_idurl):
        if not wrapped_packet or not wrapped_packet.Valid():
            lg.err('incoming Data() is not valid from supplier %r' % supplier_idurl)
            self.automat('supplier-failed', supplier_idurl=supplier_idurl)
            return
        supplier_revision = backup_control.IncomingSupplierBackupIndex(
            wrapped_packet,
            key_id=self.key_id,
            deleted_path_ids=self.files_to_be_deleted,
        )
        self.received_index_file_revision[supplier_idurl] = supplier_revision
        if _Debug:
            lg.dbg(_DebugLevel, 'received %s from %r with rev: %s' % (wrapped_packet, supplier_idurl, supplier_revision))
        self.automat('index-received', supplier_idurl=supplier_idurl)

    def _do_send_index_file(self, supplier_idurl):
        packetID = global_id.MakeGlobalID(
            key_id=self.key_id,
            path=settings.BackupIndexFileName(),
        )
        data = bpio.ReadBinaryFile(settings.BackupIndexFilePath(self.customer_idurl, self.key_alias))
        b = encrypted.Block(
            CreatorID=my_id.getIDURL(),
            BackupID=packetID,
            BlockNumber=0,
            SessionKey=key.NewSessionKey(session_key_type=key.SessionKeyType()),
            SessionKeyType=key.SessionKeyType(),
            LastBlock=True,
            Data=data,
            EncryptKey=self.key_id,
        )
        Payload = b.Serialize()
        sc = supplier_connector.by_idurl(supplier_idurl, customer_idurl=self.customer_idurl)
        if sc is None or sc.state != 'CONNECTED':
            lg.warn('supplier connector for %r is not found or offline' % supplier_idurl)
            self.automat('supplier-failed', supplier_idurl=supplier_idurl)
            return
        if online_status.isOffline(supplier_idurl):
            lg.warn('supplier %r is offline' % supplier_idurl)
            self.automat('supplier-failed', supplier_idurl=supplier_idurl)
            return
        p2p_service.SendData(
            raw_data=Payload,
            ownerID=self.customer_idurl,
            creatorID=my_id.getIDURL(),
            remoteID=supplier_idurl,
            packetID=packetID,
            callbacks={
                commands.Ack(): self._on_send_index_file_ack,
                commands.Fail(): self._on_send_index_file_ack,
            },
        )
        self.automat('index-sending', supplier_idurl=supplier_idurl)
        if _Debug:
            lg.args(_DebugLevel, pid=packetID, sz=len(data), supplier=supplier_idurl)

    def _on_supplier_failed(self, err, supplier_idurl, reason):
        lg.err('supplier %s failed with %r : %r' % (supplier_idurl, reason, err))
        self.automat('supplier-failed', supplier_idurl=supplier_idurl)
        return None

    def _on_read_customer_suppliers(self, dht_value):
        if _Debug:
            lg.args(_DebugLevel, dht_value=dht_value)
        if dht_value and isinstance(dht_value, dict) and len(dht_value.get('suppliers', [])) > 0:
            self.dht_lookup_use_cache = True
            self.automat('dht-lookup-ok', dht_value)
        else:
            self.dht_lookup_use_cache = False
            self.automat('fail', Exception('customer suppliers not found in DHT'))

    def _on_supplier_connector_state_changed(self, idurl, newstate, **kwargs):
        if _Debug:
            lg.out(_DebugLevel, 'shared_access_coordinator._supplier_connector_state_changed %s to %s, own state is %s' % (idurl, newstate, self.state))
        sc = supplier_connector.by_idurl(idurl, customer_idurl=self.customer_idurl)
        if sc:
            sc.remove_callback('shared_access_coordinator', self._on_supplier_connector_state_changed)
        if newstate == 'CONNECTED':
            self.automat('supplier-connected', supplier_idurl=idurl)

    def _on_list_files_response(self, response, info, customer_idurl, supplier_idurl, key_id):
        if _Debug:
            lg.args(_DebugLevel, response=response, customer_idurl=customer_idurl, supplier_idurl=supplier_idurl, key_id=key_id)
        self.automat('list-files-received', supplier_idurl=supplier_idurl, customer_idurl=customer_idurl, key_id=key_id)

    def _on_list_files_failed(self, response, info, customer_idurl, supplier_idurl, key_id):
        if strng.to_text(response.Payload) == 'key not registered':
            if _Debug:
                lg.dbg(_DebugLevel, 'supplier %r of customer %r do not possess public key %r yet, sending it now' % (supplier_idurl, customer_idurl, key_id))
            self.automat('key-not-registered', supplier_idurl=supplier_idurl, customer_idurl=customer_idurl, key_id=key_id)
            return None
        lg.err('failed requesting ListFiles() with %r for customer %r from supplier %r' % (key_id, customer_idurl, supplier_idurl))
        self.automat('list-files-failed', supplier_idurl=supplier_idurl, customer_idurl=customer_idurl, key_id=key_id)
        return None

    def _on_list_files_timeout(self, customer_idurl, supplier_idurl, key_id):
        lg.err('timeout requesting ListFiles() with %r for customer %r from supplier %r' % (key_id, customer_idurl, supplier_idurl))
        self.automat('list-files-failed', supplier_idurl=supplier_idurl, customer_idurl=customer_idurl, key_id=key_id)
        return None

    def _on_key_transfer_success(self, customer_idurl, supplier_idurl, key_id):
        if _Debug:
            lg.out(_DebugLevel, 'shared_access_coordinator._on_key_transfer_success public key %r shared to supplier %r of customer %r, now will send ListFiles() again' % (key_id, supplier_idurl, customer_idurl))
        self.automat('key-sent', supplier_idurl=supplier_idurl, customer_idurl=customer_idurl, key_id=key_id)

    def _on_index_file_response(self, newpacket, info):
        wrapped_packet = signed.Unserialize(newpacket.Payload)
        supplier_idurl = wrapped_packet.RemoteID if wrapped_packet else newpacket.CreatorID
        if _Debug:
            lg.args(_DebugLevel, sz=len(newpacket.Payload), s=supplier_idurl, i=info)
        if not wrapped_packet:
            lg.err('incoming Data() is not valid %r' % newpacket)
            self.automat('supplier-failed', supplier_idurl=supplier_idurl)
            return
        if not identitycache.HasKey(wrapped_packet.CreatorID):
            if _Debug:
                lg.dbg(_DebugLevel, ' will cache remote identity %s before processing incoming packet %s' % (wrapped_packet.CreatorID, wrapped_packet))
            d = identitycache.immediatelyCaching(wrapped_packet.CreatorID)
            d.addCallback(lambda _: self._do_process_index_file(wrapped_packet, supplier_idurl))
            d.addErrback(lambda err: lg.err('failed caching remote %s identity: %s' % (wrapped_packet.CreatorID, str(err))) and self.automat('supplier-failed', supplier_idurl=supplier_idurl))
            return
        self._do_process_index_file(wrapped_packet, supplier_idurl)

    def _on_index_file_fail_received(self, newpacket, info):
        if _Debug:
            lg.args(_DebugLevel, newpacket=newpacket)
        supplier_idurl = newpacket.CreatorID
        self.received_index_file_revision[supplier_idurl] = None
        if strng.to_text(newpacket.Payload) == 'key not registered':
            if _Debug:
                lg.dbg(_DebugLevel, 'supplier %r of customer %r do not possess public key %r yet, sending it now' % (supplier_idurl, self.customer_idurl, self.key_id))
            self.automat('key-not-registered', supplier_idurl=supplier_idurl, customer_idurl=self.customer_idurl, key_id=self.key_id)
            return None
        if _Debug:
            lg.out(_DebugLevel, 'shared_access_coordinator._on_index_file_fail_received %s from %r' % (newpacket, supplier_idurl))
        self.automat('index-missing', supplier_idurl=supplier_idurl)

    def _on_index_file_request_failed(self, supplier_idurl, customer_idurl, packet_id):
        if _Debug:
            lg.args(_DebugLevel, s=supplier_idurl, c=customer_idurl, pid=packet_id)
        self.received_index_file_revision[supplier_idurl] = None
        self.automat('index-failed', supplier_idurl=supplier_idurl)

    def _on_send_index_file_ack(self, newpacket, info):
        supplier_idurl = newpacket.CreatorID
        sc = supplier_connector.by_idurl(supplier_idurl, customer_idurl=self.customer_idurl)
        if sc:
            sc.automat(newpacket.Command.lower(), newpacket)
        else:
            lg.warn('did not found supplier connector for %r' % supplier_idurl)
        if _Debug:
            lg.out(_DebugLevel, 'shared_access_coordinator._on_send_index_file_ack %s from %r' % (newpacket, supplier_idurl))
        if newpacket.Command == commands.Ack():
            self.automat('index-sent', supplier_idurl=supplier_idurl)
        else:
            self.automat('index-failed', supplier_idurl=supplier_idurl)



#!/usr/bin/env python
# protocol.py
#
# Copyright (C) 2007-2008 Francois Aucamp, Meraka Institute, CSIR
# See AUTHORS for all authors and contact information.
#
# License: GNU Lesser General Public License, version 3 or later; see COPYING
#          included in this archive for details.
#
# This library is free software, distributed under the terms of
# the GNU Lesser General Public License Version 3, or any later version.
# See the COPYING file included in this archive
#
# The docstrings in this module contain epytext markup; API documentation
# may be created by processing this file with epydoc: http://epydoc.sf.net

from __future__ import absolute_import
from __future__ import print_function
import six
import time

from twisted.internet import protocol, defer
from twisted.python import failure

import twisted.internet.reactor

reactor = twisted.internet.reactor

from . import constants  # @UnresolvedImport
from . import encoding  # @UnresolvedImport
from . import msgtypes  # @UnresolvedImport
from . import msgformat  # @UnresolvedImport
from .contact import Contact, LayeredContact  # @UnresolvedImport

#------------------------------------------------------------------------------

_Debug = False

#------------------------------------------------------------------------------


class TimeoutError(Exception):
    """
    Raised when a RPC times out.
    """


class KademliaProtocol(protocol.DatagramProtocol):
    """
    Implements all low-level network-related functions of a Kademlia node.
    """
    msgSizeLimit = constants.udpDatagramMaxSize - 26
    maxToSendDelay = 10**-3
    minToSendDelay = 10**-5

    def __init__(self, node, msgEncoder=encoding.Bencode(), msgTranslator=msgformat.DefaultFormat()):
        self._node = node
        self._encoder = msgEncoder
        self._translator = msgTranslator
        self._counter = None
        self._sentMessages = {}
        self._partialMessages = {}
        self._partialMessagesProgress = {}

    def sendRPC(self, contact, method, args, rawResponse=False, **kwargs):
        """
        Sends an RPC to the specified contact.

        @param contact: The contact (remote node) to send the RPC to
        @type contact: kademlia.contacts.Contact
        @param method: The name of remote method to invoke
        @type method: str
        @param args: A list of (non-keyword) arguments to pass to the remote
                    method, in the correct order
        @type args: tuple
        @param rawResponse: If this is set to C{True}, the caller of this RPC
                            will receive a tuple containing the actual response
                            message object and the originating address tuple as
                            a result; in other words, it will not be
                            interpreted by this class. Unless something special
                            needs to be done with the metadata associated with
                            the message, this should remain C{False}.
        @type rawResponse: bool

        @return: This immediately returns a deferred object, which will return
                 the result of the RPC call, or raise the relevant exception
                 if the remote node raised one. If C{rawResponse} is set to
                 C{True}, however, it will always return the actual response
                 message (which may be a C{ResponseMessage} or an
                 C{ErrorMessage}).
        @rtype: twisted.internet.defer.Deferred
        """
        msg = msgtypes.RequestMessage(self._node.id, method, args)
        msgPrimitive = self._translator.toPrimitive(msg)
        encodedMsg = self._encoder.encode(msgPrimitive, encoding='utf-8')

        df = defer.Deferred()
        if rawResponse:
            df._rpcRawResponse = True

        # Transmit the data
        if _Debug:
            print('[DHT PROTO] <<< [%s] sendRPC' % time.time(), (method, msg.id, contact.address, contact.port, kwargs))
        if self._counter:
            self._counter('sendRPC')
        # Set the RPC timeout timer
        timeoutCall = reactor.callLater(constants.rpcTimeout*3, self._msgTimeout, msg.id)  # IGNORE:E1101
        self._sentMessages[msg.id] = (contact.id, df, timeoutCall)
        self._send(encodedMsg, msg.id, (contact.address, contact.port))
        return df

    def dispatch(self, datagram, address):
        if _Debug:
            print('[DHT PROTO]             datagram of %d bytes to dispatch from %r' % (len(datagram), address))
        msgPrimitive = self._encoder.decode(datagram, encoding='utf-8')
        if _Debug:
            print('[DHT PROTO]         msgPrimitive: %r' % msgPrimitive)
        message = self._translator.fromPrimitive(msgPrimitive)

        remoteContact = Contact(encoding.to_text(message.nodeID), address[0], address[1], self)
        # Refresh the remote node's details in the local node's k-buckets
        self._node.addContact(remoteContact)

        if _Debug:
            print('[DHT PROTO] >>> [%s] dht.dispatch %r from %r' % (
                time.time(),
                message.id,
                address,
            ))

        if isinstance(message, msgtypes.RequestMessage):
            # This is an RPC method request
            message_request = message.request
            if isinstance(message_request, six.binary_type):
                message_request = message_request.decode()
            self._handleRPC(remoteContact, message.id, message_request, message.args)
            if message.id in self._sentMessages:
                if _Debug:
                    print('[DHT PROTO]             RPC Request message received [%s]' % message_request)
                # Cancel timeout timer for this RPC
                df, timeoutCall = self._sentMessages[message.id][1:3]
                timeoutCall.cancel()
                del self._sentMessages[message.id]
            else:
                if _Debug:
                    print('[DHT PROTO]             RPC Request message %r was not identified, recently sent: %d' % (
                        message.id,
                        len([k for k in self._sentMessages.keys()]),
                    ))

        elif isinstance(message, msgtypes.ResponseMessage):
            message_response = message.response
            if isinstance(message_response, six.binary_type):
                message_response = message_response.decode()
            # Find the message that triggered this response
            if message.id in self._sentMessages:
                # Cancel timeout timer for this RPC
                df, timeoutCall = self._sentMessages[message.id][1:3]
                timeoutCall.cancel()
                del self._sentMessages[message.id]

                if hasattr(df, '_rpcRawResponse'):
                    if _Debug:
                        print('[DHT PROTO]          sent  rpcRawResponse (%r, %r)' % (message, address))
                    # The RPC requested that the raw response message and originating address be returned; do not interpret it
                    df.callback((message, address))
                elif isinstance(message, msgtypes.ErrorMessage):
                    # The RPC request raised a remote exception; raise it locally
                    remoteException = None
                    exc_msg = message_response
                    remoteException = Exception(exc_msg)
                    if _Debug:
                        print('[DHT PROTO]         sent   ErrorMessage "%s"' % exc_msg)
                    df.errback(remoteException)
                else:
                    # We got a result from the RPC
                    if _Debug:
                        print('[DHT PROTO]         sent     message response: %r' % message_response)
                    df.callback(message_response)
            else:
                # If the original message isn't found, it must have timed out
                # TODO: we should probably do something with this...
                if _Debug:
                    print('[DHT PROTO]             SKIP    message %r was not identified, recently sent: %d' % (
                        message.id,
                        len([k for k in self._sentMessages.keys()]),
                    ))
        return True

    def datagramReceived(self, datagram, address):
        """
        Handles and parses incoming RPC messages (and responses)

        @note: This is automatically called by Twisted when the protocol
               receives a UDP datagram
        """
        try:
            if _Debug:
                _t = time.time()
            # we must consistently rely on "pagination" logic actually ( or not rely at all )
            # we can't just check those two bytes in the header and say that packet is "paginated"!
            # what if a small data packet accidentally have those bytes set to \x00 just randomly?
            # so I change the protocol so it will always include such header.
            # if those two bytes are not set - it is a data coming from "late and not updated" node and we must reject it
            header_ok = False
            if datagram[0:1] == b'\x00' and datagram[45:46] == b'\x00':
                header_ok = True
            if not header_ok:
                if _Debug:
                    print('[DHT PROTO]  WARNING, dispatching old-style datagram, remote use is running old version')
                self.dispatch(datagram, address)
                return

            header = datagram[0:46]
            totalPackets = (ord(encoding.to_text(header[1:2])) << 8) | ord(encoding.to_text(header[2:3]))
            seqNumber = (ord(encoding.to_text(header[3:4])) << 8) | ord(encoding.to_text(header[4:5]))
            msgID = encoding.to_text(header[5:45], encoding='utf-8')

            if _Debug:
                print('[DHT PROTO]     datagramReceived with %d bytes   totalPackets=%d seqNumber=%d msgID=%r from %r' % (len(datagram), totalPackets, seqNumber, msgID, address))

            if seqNumber < 0 or seqNumber >= totalPackets:
                if _Debug:
                    print('[DHT PROTO]    skip, seqNumber with totalPackets')
                return

            if msgID not in self._partialMessages:
                self._partialMessages[msgID] = {}
            self._partialMessages[msgID][seqNumber] = datagram[46:]

            if len(self._partialMessages[msgID]) < totalPackets:
                if _Debug:
                    print('[DHT PROTO]   skip, _partialMessages=%r' % self._partialMessages)
                return

            keys = sorted(self._partialMessages[msgID].keys())
            data = b''
            for key in keys:
                data += self._partialMessages[msgID][key]
            datagram = data
            if _Debug:
                print('[DHT PROTO]      finished message of %d pieces: %r' % (totalPackets, keys))
            del self._partialMessages[msgID]

            self.dispatch(datagram, address)
        except Exception as exc:
            print('[DHT PROTO]       datagramReceived error:', exc)

    def _send(self, data, rpcID, address):
        """
        Transmit the specified data over UDP, breaking it up into several
        packets if necessary.

        If the data is spread over multiple UDP datagrams, the packets have the
        following structure::
            |           |     |      |      |        ||||||||||||   0x00   |
            |Transmision|Total number|Sequence number| RPC ID   |Header end|
            | type ID   | of packets |of this packet |          | indicator|
            | (1 byte)  | (2 bytes)  |  (2 bytes)    |(40 bytes)| (1 byte) |
            |           |     |      |      |        ||||||||||||          |

        @note: The header used for breaking up large data segments will
               possibly be moved out of the KademliaProtocol class in the
               future, into something similar to a message translator/encoder
               class (see C{kademlia.msgformat} and C{kademlia.encoding}).
        """
        try:
            if data is None:
                print('[DHT PROTO]       SENDING EMPTY DATA to %r at %r' % (
                    rpcID,
                    address,
                ))
                data = b''
            if len(data) > self.msgSizeLimit:
                # We have to spread the data over multiple UDP datagrams, and provide sequencing information
                # 1st byte is transmission type id, bytes 2 & 3 are the total number of packets in this transmission,
                # bytes 4 & 5 are the sequence number for this specific packet
                totalPackets = int(len(data)/self.msgSizeLimit)
                if len(data) % self.msgSizeLimit > 0:
                    totalPackets += 1
                encTotalPackets = chr(totalPackets >> 8) + chr(totalPackets & 0xff)
                seqNumber = 0
                startPos = 0
                while seqNumber < totalPackets:
                    # reactor.iterate() #IGNORE:E1101
                    packetData = data[startPos:startPos + self.msgSizeLimit]
                    encSeqNumber = chr(seqNumber >> 8) + chr(seqNumber & 0xff)
                    # actually we must always pass a header!
                    txHeader = encoding.to_bin(encTotalPackets) + encoding.to_bin(encSeqNumber) + encoding.to_bin(rpcID)
                    txData = b'\x00' + txHeader + b'\x00' + packetData
                    reactor.callLater(self.maxToSendDelay*seqNumber + self.minToSendDelay, self._write, txData, address)  # IGNORE:E1101
                    startPos += self.msgSizeLimit
                    seqNumber += 1
            else:
                totalPackets = 1
                encTotalPackets = chr(totalPackets >> 8) + chr(totalPackets & 0xff)
                seqNumber = 0
                encSeqNumber = chr(seqNumber >> 8) + chr(seqNumber & 0xff)
                txHeader = encoding.to_bin(encTotalPackets) + encoding.to_bin(encSeqNumber) + encoding.to_bin(rpcID)
                txData = b'\x00' + txHeader + b'\x00' + data
                self._write(txData, address)
        except Exception as exc:
            print('[DHT PROTO]         _send failed with: %r' % exc)

    def _write(self, data, address):
        if self._counter:
            self._counter('_write')
        try:
            self.transport.write(data, address)
        except Exception as exc:
            if _Debug:
                print('[DHT PROTO] ERROR sending UDP datagram: %r' % exc)
            return False
        if _Debug:
            print('[DHT PROTO]                     dht._write %d bytes to %s' % (len(data), str(address)))
        return True

    def _sendResponse(self, contact, rpcID, response):
        """
        Send a RPC response to the specified contact.
        """
        msg = msgtypes.ResponseMessage(rpcID, self._node.id, response)
        msgPrimitive = self._translator.toPrimitive(msg)
        encodedMsg = self._encoder.encode(msgPrimitive, encoding='utf-8')
        if _Debug:
            print('[DHT PROTO]             _sendResponse', (contact.address, contact.port), rpcID, response)
        if self._counter:
            self._counter('_sendResponse')
        self._send(encodedMsg, rpcID, (contact.address, contact.port))

    def _sendError(self, contact, rpcID, exceptionType, exceptionMessage):
        """
        Send an RPC error message to the specified contact.
        """
        msg = msgtypes.ErrorMessage(rpcID, self._node.id, exceptionType, exceptionMessage)
        msgPrimitive = self._translator.toPrimitive(msg)
        encodedMsg = self._encoder.encode(msgPrimitive, encoding='utf-8')
        if _Debug:
            print('[DHT PROTO]             _sendError', (contact.address, contact.port), rpcID, exceptionType, exceptionMessage)
        if self._counter:
            self._counter('_sendError')
        self._send(encodedMsg, rpcID, (contact.address, contact.port))

    def _handleRPC(self, senderContact, rpcID, method, args):
        """
        Executes a local function in response to an RPC request.
        """

        # Set up the deferred callchain
        def handleError(f):
            self._sendError(senderContact, rpcID, f.type, f.getErrorMessage())

        def handleResult(result):
            self._sendResponse(senderContact, rpcID, result)

        df = defer.Deferred()
        df.addCallback(handleResult)
        df.addErrback(handleError)

        if _Debug:
            print('[DHT PROTO]      _handleRPC', rpcID, method, args)

        if self._counter:
            self._counter('_handleRPC')

        # Execute the RPC
        func = getattr(self._node, method, None)

        if func and callable(func) and hasattr(func, 'rpcmethod'):
            # Call the exposed Node method and return the result to the deferred callback chain
            try:
                try:
                    # Try to pass the sender's node id to the function...
                    result = func(*args, **{'_rpcNodeID': senderContact.id, 'layerID': senderContact.layerID})
                except TypeError:
                    # import traceback
                    # traceback.print_exc()
                    # ...or simply call it if that fails
                    result = func(*args)
            except Exception as e:
                if _Debug:
                    print('[DHT PROTO]                      failed with %r' % e)
                    # import traceback
                    # traceback.print_exc()
                df.errback(failure.Failure(e))
            else:
                if _Debug:
                    print('[DHT PROTO]                   result is OK', result)
                df.callback(result)
        else:
            if _Debug:
                print('[DHT PROTO]                 no such exposed method')
            # No such exposed method
            df.errback(failure.Failure(AttributeError('Invalid method: %s' % method)))

    def _msgTimeout(self, messageID):
        """
        Called when an RPC request message times out.
        """
        if self._counter:
            self._counter('_msgTimeout')
        if _Debug:
            print(
                '[DHT PROTO]         !!! [%s] _msgTimeout' % time.time(),
                messageID,
                [k for k in self._sentMessages.keys()],
            )
        # Find the message that timed out
        if messageID in self._sentMessages:
            remoteContactID, df = self._sentMessages[messageID][0:2]
            if messageID in self._partialMessages:
                # We are still receiving this message
                # See if any progress has been made; if not, kill the message
                if messageID in self._partialMessagesProgress:
                    if len(self._partialMessagesProgress[messageID]) == len(self._partialMessages[messageID]):
                        # No progress has been made
                        del self._partialMessagesProgress[messageID]
                        del self._partialMessages[messageID]
                        df.errback(failure.Failure(TimeoutError(remoteContactID)))
                        return
                # Reset the RPC timeout timer
                timeoutCall = reactor.callLater(constants.rpcTimeout*3, self._msgTimeout, messageID)  # IGNORE:E1101
                self._sentMessages[messageID] = (remoteContactID, df, timeoutCall)
                if _Debug:
                    print('[DHT PROTO]              reset timeout for', messageID)
                return
            del self._sentMessages[messageID]
            # The message's destination node is now considered to be dead;
            # raise an (asynchronous) TimeoutError exception and update the host node
            self._node.removeContact(remoteContactID)
            df.errback(failure.Failure(TimeoutError(remoteContactID)))
        else:
            # This should never be reached
            print('[DHT PROTO]  ERROR: deferred timed out, but is not present in sent messages list!')


class KademliaMultiLayerProtocol(KademliaProtocol):
    def __init__(self, node, msgEncoder=encoding.Bencode(), msgTranslator=msgformat.MultiLayerFormat()):
        KademliaProtocol.__init__(self, node, msgEncoder=msgEncoder, msgTranslator=msgTranslator)

    def _write(self, data, address):
        self._node.bytes_out += len(data)
        return KademliaProtocol._write(self, data, address)

    def sendRPC(self, contact, method, args, rawResponse=False, layerID=0, **kwargs):
        only_ask = layerID != 0 and layerID not in self._node.attached_layers
        if only_ask:
            msg = msgtypes.QuestionMessage(self._node.layers[layerID], method, args, layerID=layerID)
        else:
            msg = msgtypes.RequestMessage(self._node.layers[layerID], method, args, layerID=layerID)
        msgPrimitive = self._translator.toPrimitive(msg)
        encodedMsg = self._encoder.encode(msgPrimitive, encoding='utf-8')
        df = defer.Deferred()
        if rawResponse:
            df._rpcRawResponse = True
        if _Debug:
            print('[DHT PROTO] <<< [%s] sendRPC  layerID=%d method=%r msg_id=%r addr=%r port=%r kwargs=%r' % (msg, layerID, method, msg.id, contact.address, contact.port, kwargs))
        if self._counter:
            self._counter('sendRPC')
        timeoutCall = reactor.callLater(constants.rpcTimeout*3, self._msgTimeout, msg.id)  # IGNORE:E1101
        self._sentMessages[msg.id] = (contact.id, df, timeoutCall, method)
        self._send(encodedMsg, msg.id, (contact.address, contact.port))
        if layerID not in self._node.packets_out:
            self._node.packets_out[layerID] = 0
        self._node.packets_out[layerID] += 1
        return df

    def dispatch(self, datagram, address):
        if _Debug:
            print('[DHT PROTO]    dispatch datagram of %d bytes to dispatch from %r' % (len(datagram), address))
        msgPrimitive = self._encoder.decode(datagram, encoding='utf-8')
        if _Debug:
            print('[DHT PROTO]                 msgPrimitive: %r' % msgPrimitive)
        message = self._translator.fromPrimitive(msgPrimitive)
        layerID = message.layerID
        if layerID not in self._node.layers:
            # TODO: add protection here
            if _Debug:
                print('[DHT PROTO]    dispatch datagram %r was not identified, layer %d not opened: %r' % (
                    message.id,
                    layerID,
                ))
            return True

        if layerID not in self._node.packets_in:
            self._node.packets_in[layerID] = 0
        self._node.packets_in[layerID] += 1

        remoteContact = LayeredContact(encoding.to_text(message.nodeID), address[0], address[1], self, layerID=layerID)

        if not isinstance(message, msgtypes.QuestionMessage):
            # Refresh the remote node's details in the local node's k-buckets
            self._node.addContact(remoteContact, layerID=layerID)
            if _Debug:
                print('[DHT PROTO]    dispatch  add or update contact %r at layer %d to my routing table' % (
                    remoteContact,
                    layerID,
                ))
        else:
            if _Debug:
                print('[DHT PROTO]    dispatch  skip adding contact %r at layer %d to my routing table' % (
                    remoteContact,
                    layerID,
                ))

        if _Debug:
            print('[DHT PROTO]    dispatch  >>> [%s] dht.dispatch %r from %r' % (
                time.time(),
                message.id,
                address,
            ))

        if isinstance(message, msgtypes.RequestMessage) or isinstance(message, msgtypes.QuestionMessage):
            # This is an RPC method request
            message_request = message.request
            if isinstance(message_request, six.binary_type):
                message_request = message_request.decode()
            self._handleRPC(remoteContact, message.id, message_request, message.args)

            if layerID not in self._node.rpc_calls:
                self._node.rpc_calls[layerID] = {}
            if message_request not in self._node.rpc_calls[layerID]:
                self._node.rpc_calls[layerID][message_request] = 0
            self._node.rpc_calls[layerID][message_request] += 1

            if message.id in self._sentMessages:
                if _Debug:
                    print('[DHT PROTO]    dispatch   RPC Request message received [%s]' % message_request)
                # Cancel timeout timer for this RPC
                df, timeoutCall = self._sentMessages[message.id][1:3]
                timeoutCall.cancel()
                del self._sentMessages[message.id]
            else:
                if _Debug:
                    print('[DHT PROTO]    dispatch   RPC Request message %r is not a reply, latest outgoing messages: %r' % (
                        message.id,
                        [k for k in self._sentMessages.keys()],
                    ))

        elif isinstance(message, msgtypes.ResponseMessage):
            message_response = message.response
            if isinstance(message_response, six.binary_type):
                message_response = message_response.decode()
            # Find the message that triggered this response
            if message.id in self._sentMessages:
                # Cancel timeout timer for this RPC
                df, timeoutCall, method = self._sentMessages[message.id][1:4]
                timeoutCall.cancel()
                del self._sentMessages[message.id]

                if layerID not in self._node.rpc_responses:
                    self._node.rpc_responses[layerID] = {}
                if method not in self._node.rpc_responses[layerID]:
                    self._node.rpc_responses[layerID][method] = 0
                self._node.rpc_responses[layerID][method] += 1

                if hasattr(df, '_rpcRawResponse'):
                    if _Debug:
                        print('[DHT PROTO]    dispatch   respond with tuple (%r, %r)' % (message, address))
                    # The RPC requested that the raw response message and originating address be returned; do not interpret it
                    df.callback((message, address))
                elif isinstance(message, msgtypes.ErrorMessage):
                    # The RPC request raised a remote exception; raise it locally
                    remoteException = None
                    exc_msg = message_response
                    remoteException = Exception(exc_msg)
                    if _Debug:
                        print('[DHT PROTO]    dispatch   respond with error "%s"' % exc_msg)
                    df.errback(remoteException)
                else:
                    # We got a result from the RPC
                    if _Debug:
                        print('[DHT PROTO]    dispatch  respond with message_response: %r' % message_response)
                    df.callback(message_response)
            else:
                if layerID not in self._node.rpc_responses:
                    self._node.rpc_responses[layerID] = {}
                if 'timeout' not in self._node.rpc_responses[layerID]:
                    self._node.rpc_responses[layerID]['timeout'] = 0
                self._node.rpc_responses[layerID]['timeout'] += 1

                # If the original message isn't found, it must have timed out
                # TODO: we should probably do something with this...
                if _Debug:
                    print('[DHT PROTO]    dispatch  message %r was not identified, currently sent: %r' % (
                        message.id,
                        [k for k in self._sentMessages.keys()],
                    ))
        return True

    def _sendResponse(self, contact, rpcID, response):
        """
        Send a RPC response to the specified contact.
        """
        layerID = contact.layerID
        msg = msgtypes.ResponseMessage(rpcID, self._node.layers[layerID], response, layerID=layerID)
        msgPrimitive = self._translator.toPrimitive(msg)
        encodedMsg = self._encoder.encode(msgPrimitive, encoding='utf-8')
        if _Debug:
            print('[DHT PROTO]             _sendResponse', (contact.address, contact.port), rpcID, response)
        if self._counter:
            self._counter('_sendResponse')
        self._send(encodedMsg, rpcID, (contact.address, contact.port))

    def _sendError(self, contact, rpcID, exceptionType, exceptionMessage):
        """
        Send an RPC error message to the specified contact.
        """
        layerID = contact.layerID
        msg = msgtypes.ErrorMessage(rpcID, self._node.layers[layerID], exceptionType, exceptionMessage, layerID=layerID)
        msgPrimitive = self._translator.toPrimitive(msg)
        encodedMsg = self._encoder.encode(msgPrimitive, encoding='utf-8')
        if _Debug:
            print('[DHT PROTO]              _sendError', (contact.address, contact.port), rpcID, exceptionType, exceptionMessage)
        if self._counter:
            self._counter('_sendError')
        self._send(encodedMsg, rpcID, (contact.address, contact.port))

    def datagramReceived(self, datagram, address):
        self._node.bytes_in += len(datagram)
        try:
            # we must consistently rely on "pagination" logic actually ( or not rely at all )
            # we can't just check those two bytes in the header and say that packet is "paginated"!
            # what if a small data packet accidentally have those bytes set to \x00 just randomly?
            # so I change the protocol so it will always include such header.
            # if those two bytes are not set - it is a data coming from "late and not updated" node and we must reject it
            header_ok = False
            if datagram[0:1] == b'\x00' and datagram[45:46] == b'\x00':
                header_ok = True
            if not header_ok:
                if _Debug:
                    print('[DHT PROTO]  WARNING, dispatching old-style datagram, remote use is running old version')
                self.dispatch(datagram, address)
                return

            header = datagram[0:46]
            totalPackets = (ord(encoding.to_text(header[1:2])) << 8) | ord(encoding.to_text(header[2:3]))
            seqNumber = (ord(encoding.to_text(header[3:4])) << 8) | ord(encoding.to_text(header[4:5]))
            msgID = encoding.to_text(header[5:45], encoding='utf-8')

            if _Debug:
                print('[DHT PROTO]     datagramReceived with %d bytes   totalPackets=%d seqNumber=%d msgID=%r from %r' % (len(datagram), totalPackets, seqNumber, msgID, address))

            if seqNumber < 0 or seqNumber >= totalPackets:
                if _Debug:
                    print('[DHT PROTO]    skip, seqNumber with totalPackets')
                return

            if msgID not in self._partialMessages:
                self._partialMessages[msgID] = {}
            self._partialMessages[msgID][seqNumber] = datagram[46:]

            if len(self._partialMessages[msgID]) < totalPackets:
                if _Debug:
                    print('[DHT PROTO]   skip, _partialMessages=%r' % self._partialMessages)
                return

            keys = sorted(self._partialMessages[msgID].keys())
            data = b''
            for key in keys:
                data += self._partialMessages[msgID][key]
            datagram = data
            if _Debug:
                print('[DHT PROTO]      finished message of %d pieces: %r' % (totalPackets, keys))
            del self._partialMessages[msgID]

            self.dispatch(datagram, address)
        except Exception as exc:
            print('[DHT PROTO]         datagramReceived error:', exc)

    def _msgTimeout(self, messageID):
        """
        Called when an RPC request message times out.
        """
        if self._counter:
            self._counter('_msgTimeout')
        if _Debug:
            print(
                '[DHT PROTO]         !!! [%s] _msgTimeout' % time.time(),
                messageID,
                [k for k in self._sentMessages.keys()],
            )
        # Find the message that timed out
        if messageID in self._sentMessages:
            remoteContactID, df, _, method = self._sentMessages[messageID][0:4]
            if messageID in self._partialMessages:
                # We are still receiving this message
                # See if any progress has been made; if not, kill the message
                if messageID in self._partialMessagesProgress:
                    if len(self._partialMessagesProgress[messageID]) == len(self._partialMessages[messageID]):
                        # No progress has been made
                        del self._partialMessagesProgress[messageID]
                        del self._partialMessages[messageID]
                        df.errback(failure.Failure(TimeoutError(remoteContactID)))
                        return
                # Reset the RPC timeout timer
                timeoutCall = reactor.callLater(constants.rpcTimeout*3, self._msgTimeout, messageID)  # IGNORE:E1101
                self._sentMessages[messageID] = (remoteContactID, df, timeoutCall, method)
                if _Debug:
                    print('[DHT PROTO]              reset timeout for', messageID)
                return
            del self._sentMessages[messageID]
            # The message's destination node is now considered to be dead;
            # raise an (asynchronous) TimeoutError exception and update the host node
            self._node.removeContact(remoteContactID)
            df.errback(failure.Failure(TimeoutError(remoteContactID)))
        else:
            # This should never be reached
            print('[DHT PROTO]  ERROR: deferred timed out, but is not present in sent messages list!')



# Save this in to a file sum_client.py

from __future__ import absolute_import
from __future__ import print_function
from twisted.internet import reactor  # @UnresolvedImport
from twisted.internet.protocol import ClientCreator
from twisted.protocols import amp


class Sum(amp.Command):
    # normally shared by client and server code
    arguments = [(b'a', amp.Integer()), (b'b', amp.Integer())]
    response = [(b'total', amp.Integer())]


def connected(protocol):
    return protocol.callRemote(Sum, a=10, b=81).addCallback(gotResult)


def gotResult(result):
    print(('total: %d' % (result['total'], )))
    reactor.stop()


def error(reason):
    print('Something went wrong')
    print(reason)
    reactor.stop()


ClientCreator(reactor, amp.AMP).connectTCP('127.0.0.1', 1234).addCallback(connected).addErrback(error)

reactor.run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2011-2013 Codernity (http://codernity.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import
import six
import functools
from random import choice
from six.moves import range


def create_cache1lvl(lock_obj):
    def cache1lvl(maxsize=100):
        def decorating_function(user_function):
            cache = {}
            lock = lock_obj()

            @functools.wraps(user_function)
            def wrapper(key, *args, **kwargs):
                if isinstance(key, six.text_type):
                    key = key.encode()
                try:
                    result = cache[key]
                except KeyError:
                    with lock:
                        if len(cache) == maxsize:
                            for i in range(maxsize // 10 or 1):
                                del cache[choice(list(cache.keys()))]
                        cache[key] = user_function(key, *args, **kwargs)
                        result = cache[key]
                return result

            def clear():
                cache.clear()

            def delete(key):
                if isinstance(key, six.text_type):
                    key = key.encode()
                try:
                    del cache[key]
                    return True
                except KeyError:
                    return False

            wrapper.clear = clear
            wrapper.cache = cache
            wrapper.delete = delete
            return wrapper
        return decorating_function
    return cache1lvl


def create_cache2lvl(lock_obj):
    def cache2lvl(maxsize=100):
        def decorating_function(user_function):
            cache = {}
            lock = lock_obj()

            @functools.wraps(user_function)
            def wrapper(*args, **kwargs):
                key = args[0]
                # print('rr_cache_with_lock.2lvl.wrapper %r' % key)
                if isinstance(key, six.text_type):
                    key = key.encode()
                try:
                    result = cache[key][args[1]]
                except KeyError:
                    with lock:
                        if wrapper.cache_size == maxsize:
                            to_delete = maxsize // 10 or 1
                            for i in range(to_delete):
                                key1 = choice(list(cache.keys()))
                                key2 = choice(list(cache[key1].keys()))
                                del cache[key1][key2]
                                if not cache[key1]:
                                    del cache[key1]
                            wrapper.cache_size -= to_delete
                        result = user_function(*args, **kwargs)
                        try:
                            cache[key][args[1]] = result
                        except KeyError:
                            cache[key] = {args[1]: result}
                        wrapper.cache_size += 1
                return result

            def clear():
                cache.clear()
                wrapper.cache_size = 0

            def delete(key, *args):
                # print('rr_cache_with_lock.2lvl.delete %r' % key)
                if isinstance(key, six.text_type):
                    key = key.encode()
                if args:
                    try:
                        del cache[key][args[0]]
                        if not cache[key]:
                            del cache[key]
                        wrapper.cache_size -= 1
                        return True
                    except KeyError:
                        return False
                else:
                    try:
                        wrapper.cache_size -= len(cache[key])
                        del cache[key]
                        return True
                    except KeyError:
                        return False

            wrapper.clear = clear
            wrapper.cache = cache
            wrapper.delete = delete
            wrapper.cache_size = 0
            return wrapper
        return decorating_function
    return cache2lvl



#!/usr/bin/env python
# p2p_connector.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (p2p_connector.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
"""
.. module:: p2p_connector.

.. raw:: html

    <a href="https://bitdust.io/automats/p2p_connector/p2p_connector.png" target="_blank">
    <img src="https://bitdust.io/automats/p2p_connector/p2p_connector.png" style="max-width:100%;">
    </a>

The ``p2p_connector()`` state machine manages the user's connection with other remote users.

It calls other state machines and works with them in parallel.

Control passes to the ``network_connector()`` automat which prepares the Internet connection.

Next, there is a start of transport protocols.

User Identity file that contains a public user address get updated.
The new version of the file is sent to the Identity server
in order for that to all the other users can find out the contact with user.

Then, as soon as user receive the first packet from any of the remote users,
the ``p2p_connector()`` goes into state "CONNECTED".

If user changes his network settings all process should be restarted.

EVENTS:
    * :red:`check-synchronize`
    * :red:`inbox-packet`
    * :red:`init`
    * :red:`instant`
    * :red:`my-id-propagated`
    * :red:`my-id-updated`
    * :red:`network_connector.state`
    * :red:`ping-contact`
    * :red:`timer-20sec`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from six.moves import range

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 16

#------------------------------------------------------------------------------

import time

from twisted.internet.task import LoopingCall  #@UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio

from bitdust.main import settings
from bitdust.main import config
from bitdust.main import events

from bitdust.lib import net_misc
from bitdust.lib import strng

from bitdust.automats import automat

from bitdust.contacts import identitycache
from bitdust.contacts import contactsdb

from bitdust.transport import callback

from bitdust.services import driver

from bitdust.p2p import propagate
from bitdust.p2p import network_connector

from bitdust.userid import identity
from bitdust.userid import my_id

#------------------------------------------------------------------------------

_P2PConnector = None
_ActiveProtocols = set()

#------------------------------------------------------------------------------


def active_protos():
    global _ActiveProtocols
    return _ActiveProtocols


def inbox(newpacket, info, status, message):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_connector.inbox %s %s' % (status, newpacket))
    # here need to mark this protocol as working
    if info.proto in [
        'tcp',
    ]:
        if not net_misc.IpIsLocal(strng.to_text(info.host).split(':')[0]):
            # but we want to check that this packet is come from the Internet, not our local network
            # because we do not want to use this proto as first method if it is not working for all
            if info.proto not in active_protos():
                if _Debug:
                    lg.out(_DebugLevel - 6, 'p2p_connector.inbox [transport_%s] seems to work !!!!!!!!!!!!!!!!!!!!!' % info.proto)
                    lg.out(_DebugLevel - 6, '                    We got packet from %s://%s' % (info.proto, str(info.host)))
                active_protos().add(info.proto)
    elif info.proto in [
        'http',
    ]:
        if not net_misc.IpIsLocal(strng.to_text(info.host).split(':')[0]):
            # but we want to check that this packet is come from the Internet, not our local network
            # because we do not want to use this proto as first method if it is not working for all
            if info.proto not in active_protos():
                if _Debug:
                    lg.out(_DebugLevel - 6, 'p2p_connector.inbox [transport_%s] seems to work !!!!!!!!!!!!!!!!!!!!!' % info.proto)
                    lg.out(_DebugLevel - 6, '                    We got packet from %s://%s' % (info.proto, str(info.host)))
                active_protos().add(info.proto)
    elif info.proto in [
        'udp',
    ]:
        if info.proto not in active_protos():
            if _Debug:
                lg.out(_DebugLevel - 6, 'p2p_connector.inbox [transport_%s] seems to work !!!!!!!!!!!!!!!!!!!!!' % info.proto)
                lg.out(_DebugLevel - 6, '                    We got packet from %s://%s' % (info.proto, str(info.host)))
            active_protos().add(info.proto)
    elif info.proto in [
        'proxy',
    ]:
        if info.proto not in active_protos():
            if _Debug:
                lg.out(_DebugLevel - 6, 'p2p_connector.inbox [transport_%s] seems to work !!!!!!!!!!!!!!!!!!!!!' % info.proto)
                lg.out(_DebugLevel - 6, '                    We got packet from %s://%s' % (info.proto, str(info.host)))
            active_protos().add(info.proto)
    A('inbox-packet', (newpacket, info, status, message))
    return False


#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _P2PConnector
    if event is None:
        return _P2PConnector
    if _P2PConnector is None:
        _P2PConnector = P2PConnector(
            name='p2p_connector',
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=False,
            log_transitions=_Debug,
        )
    if event is not None:
        _P2PConnector.automat(event, *args, **kwargs)
    return _P2PConnector


def Destroy():
    """
    Destroy p2p_connector() automat and remove its instance from memory.
    """
    global _P2PConnector
    if _P2PConnector is None:
        return
    _P2PConnector.destroy()
    del _P2PConnector
    _P2PConnector = None


class P2PConnector(automat.Automat):

    fast = False

    timers = {
        'timer-20sec': (20.0, ['INCOMMING?']),
    }

    def init(self):
        self.is_reconnecting = False
        self.health_check_task = None

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        if oldstate != newstate and oldstate == 'MY_IDENTITY':
            self.is_reconnecting = False
        if newstate == 'INCOMMING?' and event != 'instant':
            self.automat('instant')
        if newstate == 'CONNECTED':
            if self.health_check_task:
                self.health_check_task.stop()
            health_check_interval = config.conf().getInt('services/identity-propagate/health-check-interval-seconds')
            if health_check_interval:
                self.health_check_task = LoopingCall(self._do_id_server_health_check)
                self.health_check_task.start(health_check_interval, now=False)
        else:
            if self.health_check_task:
                self.health_check_task.stop()
                self.health_check_task = None

    def A(self, event, *args, **kwargs):
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'NETWORK?'
                self.NeedPropagate = True
                self.doInit(*args, **kwargs)
                network_connector.A('reconnect')
        #---NETWORK?---
        elif self.state == 'NETWORK?':
            if (event == 'network_connector.state' and args[0] == 'CONNECTED'):
                self.state = 'MY_IDENTITY'
                self.doUpdateMyIdentity(event, *args, **kwargs)
            elif (event == 'network_connector.state' and args[0] == 'DISCONNECTED'):
                self.state = 'DISCONNECTED'
        #---INCOMMING?---
        elif self.state == 'INCOMMING?':
            if event == 'inbox-packet' and not self.isUsingBestProto(*args, **kwargs):
                self.state = 'MY_IDENTITY'
                self.doUpdateMyIdentity(event, *args, **kwargs)
                self.doPopBestProto(*args, **kwargs)
            elif (event == 'instant' and not self.isAnyPeersKnown(*args, **kwargs)) or (event == 'inbox-packet' and self.isUsingBestProto(*args, **kwargs)):
                self.state = 'CONNECTED'
                self.doInitRatings(*args, **kwargs)
                self.doRestartCustomersRejector(*args, **kwargs)
            elif event == 'timer-20sec' or (event == 'network_connector.state' and args[0] == 'DISCONNECTED'):
                self.state = 'DISCONNECTED'
                self.doInitRatings(*args, **kwargs)
            elif event == 'check-synchronize' or (event == 'network_connector.state' and args[0] == 'CONNECTED'):
                self.state = 'MY_IDENTITY'
                self.doUpdateMyIdentity(event, *args, **kwargs)
        #---CONNECTED---
        elif self.state == 'CONNECTED':
            if event == 'ping-contact':
                self.doSendMyIdentity(*args, **kwargs)
            elif (event == 'network_connector.state' and args[0] not in ['CONNECTED', 'DISCONNECTED']):
                self.state = 'NETWORK?'
            elif (event == 'network_connector.state' and args[0] == 'DISCONNECTED'):
                self.state = 'DISCONNECTED'
            elif event == 'check-synchronize' or (event == 'network_connector.state' and args[0] == 'CONNECTED'):
                self.state = 'MY_IDENTITY'
                self.doUpdateMyIdentity(event, *args, **kwargs)
        #---DISCONNECTED---
        elif self.state == 'DISCONNECTED':
            if event == 'inbox-packet' or event == 'check-synchronize' or ((event == 'network_connector.state' and args[0] == 'CONNECTED')):
                self.state = 'MY_IDENTITY'
                self.doUpdateMyIdentity(event, *args, **kwargs)
            elif event == 'ping-contact':
                self.doSendMyIdentity(*args, **kwargs)
            elif (event == 'network_connector.state' and args[0] not in ['CONNECTED', 'DISCONNECTED']):
                self.state = 'NETWORK?'
        #---MY_IDENTITY---
        elif self.state == 'MY_IDENTITY':
            if event == 'my-id-updated' and self.isMyContactsChanged(*args, **kwargs):
                self.state = 'NETWORK?'
                self.NeedPropagate = True
                network_connector.A('check-reconnect')
            elif event == 'my-id-updated' and not self.isMyContactsChanged(*args, **kwargs) and (self.NeedPropagate or self.isMyIdentityChanged(*args, **kwargs)):
                self.state = 'PROPAGATE'
                self.doCheckRotatePropagateMyIdentity(*args, **kwargs)
            elif event == 'my-id-updated' and not (self.NeedPropagate or self.isMyIdentityChanged(*args, **kwargs)) and (network_connector.A().state == 'CONNECTED'):
                self.state = 'CONNECTED'
            elif event == 'my-id-updated' and not (self.NeedPropagate or self.isMyIdentityChanged(*args, **kwargs)) and (network_connector.A().state != 'CONNECTED'):
                self.state = 'DISCONNECTED'
        #---PROPAGATE---
        elif self.state == 'PROPAGATE':
            if event == 'my-id-propagated':
                self.state = 'INCOMMING?'
                self.NeedPropagate = False
                self.doRestartFireHire(*args, **kwargs)
            elif ((event == 'network_connector.state' and args[0] == 'CONNECTED')) or event == 'check-synchronize':
                self.state = 'MY_IDENTITY'
                self.doUpdateMyIdentity(event, *args, **kwargs)
        return None

    def isUsingBestProto(self, *args, **kwargs):
        """
        Condition method.
        """
        return self._check_to_use_best_proto()

    def isMyIdentityChanged(self, *args, **kwargs):
        """
        Condition method.
        """
        return args[0][1]

    def isMyContactsChanged(self, *args, **kwargs):
        """
        Condition method.
        """
        if self.is_reconnecting:
            return True
        return args[0][0]

    def isAnyPeersKnown(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(list(filter(None, contactsdb.contacts_remote()))) > 0

    def doInit(self, *args, **kwargs):
        version_number = bpio.ReadTextFile(settings.VersionNumberFile()).strip()
        if _Debug:
            lg.out(_DebugLevel - 6, 'p2p_connector.doInit RevisionNumber=%s' % str(version_number))
        callback.append_inbox_callback(inbox)

    def doSendMyIdentity(self, *args, **kwargs):
        """
        Action method.
        """
        propagate.single(args[0], wide=True)

    def doUpdateMyIdentity(self, event, *args, **kwargs):
        if _Debug:
            lg.out(_DebugLevel - 6, 'p2p_connector.doUpdateMyIdentity event=%r' % event)
        if event == 'check-synchronize':
            self.is_reconnecting = True
        self._update_my_identity()

    def doCheckRotatePropagateMyIdentity(self, *args, **kwargs):
        if _Debug:
            lg.out(_DebugLevel - 6, 'p2p_connector.doCheckRotatePropagateMyIdentity')
        self._check_rotate_propagate_my_identity()

    def doPopBestProto(self, *args, **kwargs):
        self._pop_active_proto()

    def doInitRatings(self, *args, **kwargs):
        """
        Action method.
        """
        # ratings.init()

    def doRestartCustomersRejector(self, *args, **kwargs):
        """
        Action method.
        """
        if driver.is_on('service_customer_patrol'):
            # TODO: move this into a callback inside service_customer_patrol
            from bitdust.supplier import customers_rejector
            customers_rejector.A('restart')

    def doRestartFireHire(self, *args, **kwargs):
        """
        Action method.
        """
        if driver.is_on('service_employer'):
            # TODO: use events instead
            from bitdust.customer import fire_hire
            fire_hire.A('restart')

    def _check_to_use_best_proto(self):
        # if no incoming traffic - do nothing
        if len(active_protos()) == 0:
            return True
        lid = my_id.getLocalIdentity()
        order = lid.getProtoOrder()
        # if no protocols in local identity - do nothing
        if len(order) == 0:
            return True
        # when transport proxy is working we do not need to check our contacts at all
        if settings.transportIsEnabled('proxy'):
            if driver.is_on('service_proxy_transport'):
                if settings.transportReceivingIsEnabled('proxy'):
                    try:
                        # TODO: change here to receive the value directly from service_proxy_transport object
                        router_idurl = driver.services()['service_proxy_transport'].transport.options['router_idurl']
                    except:
                        router_idurl = None
                    if router_idurl:
                        router_identity = identitycache.FromCache(router_idurl)
                        contacts_is_ok = True
                        router_protos = router_identity.getContactsByProto()
                        if lid.getContactsNumber() != len(router_protos):
                            contacts_is_ok = False
                        if contacts_is_ok:
                            for proto, contact in router_protos.items():
                                if lid.getProtoContact(proto) != contact:
                                    contacts_is_ok = False
                        if contacts_is_ok:
                            if _Debug:
                                lg.out(_DebugLevel - 6, 'p2p_connector._check_to_use_best_proto returning True : proxy_transport is OK')
                            return True
        first = order[0]
        # if first contact in local identity is not working yet
        # but there is another working methods - switch first method
        if first not in active_protos():
            if _Debug:
                lg.out(_DebugLevel - 6, 'p2p_connector._check_to_use_best_proto first contact (%s) is not working!   active_protos()=%s' % (first, str(active_protos())))
            return False
        # #small hack to make udp as first method if all is fine
        # if first != 'udp' and ('udp' in active_protos() and 'tcp' in active_protos()):
        #     lg.out(2, 'p2p_connector._check_to_use_best_proto first contact (%s) but UDP also works!  active_protos()=%s' % (first, str(active_protos())))
        #     return False
        # if tcp contact is on first place and it is working - we are VERY HAPPY! - no need to change anything - return False
        if first == 'tcp' and 'tcp' in active_protos():
            return True
        # but if tcp method is not the first and it works - we want to TURN IT ON! - return True
        if first != 'tcp' and 'tcp' in active_protos():
            if _Debug:
                lg.out(_DebugLevel - 6, 'p2p_connector._check_to_use_best_proto tcp is not first but it works active_protos()=%s' % str(active_protos()))
            return False
        # if we are using udp and it is working - this is fantastic!
        if first == 'udp' and 'udp' in active_protos():
            # but let's check if TCP is also working
            # in that case we want to switch to TCP
            if 'tcp' in active_protos():
                return False
            return True
        # udp seems to be working and first contact is not working - so switch to udp
        if first != 'udp' and 'udp' in active_protos():
            if _Debug:
                lg.out(_DebugLevel - 6, 'p2p_connector._check_to_use_best_proto udp is not first but it works active_protos()=%s' % str(active_protos()))
            return False
        # http seems to work and it is first - cool!
        if first == 'http' and 'http' in active_protos():
            return True
        # but if http method is not the first and it works - we want to TURN IT ON! - return True
        if first != 'http' and 'http' in active_protos():
            if _Debug:
                lg.out(_DebugLevel - 6, 'p2p_connector._check_to_use_best_proto http is not first but it works active_protos()=%s' % str(active_protos()))
            return False
        # if we are using proxy and it is working - that is fine - it must work always!
        if first == 'proxy' and 'proxy' in active_protos():
            return True
        # proxy seems to be working and first contact is not working - so switch to proxy
        if first != 'proxy' and 'proxy' in active_protos():
            if _Debug:
                lg.out(_DebugLevel - 6, 'p2p_connector._check_to_use_best_proto proxy is not first but it works active_protos()=%s' % str(active_protos()))
            return False
        # in other cases - do nothing
        return True

    def _pop_active_proto(self):
        if len(active_protos()) == 0:
            return
        lid = my_id.getLocalIdentity()
        order = lid.getProtoOrder()
        first = order[0]
        wantedproto = ''
        # if first contact in local identity is not working yet
        # but there is another working methods - switch first method
        if first not in active_protos():
            # take (but not remove) any item from the set
            wantedproto = active_protos().pop()
            active_protos().add(wantedproto)
        # if proxy method is not the first but it works - switch to proxy
        if first != 'proxy' and 'proxy' in active_protos():
            wantedproto = 'proxy'
        # if http method is not the first but it works - switch to http
        if first != 'http' and 'http' in active_protos():
            wantedproto = 'http'
        # if udp method is not the first but it works - switch to udp
        if first != 'udp' and 'udp' in active_protos():
            wantedproto = 'udp'
        # if tcp method is not the first but it works - switch to tcp
        if first != 'tcp' and 'tcp' in active_protos():
            wantedproto = 'tcp'
        if _Debug:
            lg.out(_DebugLevel - 6, 'p2p_connector.PopWorkingProto will pop %s contact order=%s active_protos()=%s' % (wantedproto, str(order), str(active_protos())))
        # now move best proto on the top
        # other users will use this method to send to us
        lid.popProtoContact(wantedproto)
        # save local id
        # also need to propagate our identity
        # other users must know our new contacts
        my_id.setLocalIdentity(lid)
        my_id.saveLocalIdentity()

    def _update_my_identity(self):
        contacts_changed = False
        old_contacts = list(my_id.getLocalIdentity().getContacts())
        identity_changed = my_id.rebuildLocalIdentity()
        if not identity_changed and len(active_protos()) > 0:
            if _Debug:
                lg.args(_DebugLevel, identity_changed=identity_changed, contacts_changed=contacts_changed, active_protos=active_protos())
            self.automat('my-id-updated', (False, False))
            return
        new_contacts = my_id.getLocalIdentity().getContacts()
        if len(old_contacts) != len(new_contacts):
            contacts_changed = True
        if not contacts_changed:
            for pos in range(len(old_contacts)):
                if old_contacts[pos] != new_contacts[pos]:
                    contacts_changed = True
                    break
        if contacts_changed:
            # erase all stats about received packets
            # if some contacts in my identity has been changed
            if _Debug:
                lg.out(4, '    my contacts were changed, erasing active_protos() flags')
            active_protos().clear()
        if _Debug:
            lg.out(4, '    identity HAS %sBEEN CHANGED' % ('' if identity_changed else 'NOT '))
        self.automat('my-id-updated', (contacts_changed, identity_changed))

    def _do_propagate(self, result, rotated):
        if _Debug:
            lg.out(_DebugLevel, 'p2p_connector._do_propagate rotated=%r result=%r' % (rotated, result))
        if driver.is_on('service_entangled_dht'):
            from bitdust.dht import dht_service
            dht_service.set_node_data('idurl', my_id.getIDURL().to_text())
        else:
            lg.warn('DHT node local store was not updated with my latest IDURL, service_entangled_dht was OFF')
        d = propagate.start(
            wide=True,
            refresh_cache=True,
            include_all=True,
            include_startup=rotated,
            wait_packets=True,
        )
        d.addCallback(lambda x: self.automat('my-id-propagated'))
        if rotated:
            d.addBoth(lambda x: events.send('my-identity-rotate-complete', data=dict()))
        d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='_check_rotate_propagate_my_identity._do_propagate')
        d.addErrback(self._on_propagate_failed)
        # d.addTimeout(30, clock=reactor)
        return result

    def _on_propagate_failed(self, err):
        lg.err('failed propagate my identity: %r' % err)
        self.automat('my-id-propagated')
        return err

    def _on_update_failed(self, err):
        lg.err('failed to update my identity: %r' % err)
        self.automat('my-id-propagated')
        return err

    def _do_update(self, ret):
        if _Debug:
            lg.out(_DebugLevel, 'p2p_connector._do_update  ret=%r' % repr(ret))
        result, rotated = ret
        if not result:
            lg.err('failed to rotate identity sources, skip propagating my identity')
            self.automat('my-id-propagated')
            return None
        d = propagate.update()
        d.addCallback(self._do_propagate, rotated)
        d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='_check_rotate_propagate_my_identity._do_update')
        d.addErrback(self._on_update_failed)
        return ret

    def _do_rotate(self, ret):
        if _Debug:
            lg.out(_DebugLevel, 'p2p_connector._do_rotate  ret=%r' % repr(ret))
        check_result, rotated = ret
        if check_result:
            lg.info('identity sources are healthy, send my identity now')
            self._do_update((
                True,
                False,
            ))
            return None
        lg.err('identity sources are not healthy, will execute identity rotate flow now')
        from bitdust.p2p import id_rotator
        d = id_rotator.run()
        d.addCallback(self._do_update)
        d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='_check_rotate_propagate_my_identity._do_rotate')
        d.addErrback(lambda _: self._do_update((False, False)))
        return ret

    def _do_check(self, ret):
        if _Debug:
            lg.out(_DebugLevel, 'p2p_connector._do_check ret=%r' % repr(ret))
        from bitdust.p2p import id_rotator
        d = id_rotator.check()
        d.addCallback(self._do_rotate)
        d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='_check_rotate_propagate_my_identity._do_check')
        d.addErrback(lambda _: self._do_rotate((False, False)))
        return ret

    def _check_rotate_propagate_my_identity(self):
        # TODO: rebuild that method into another state machine
        self._do_check(None)

    def _do_id_server_health_check(self):
        my_idurl = my_id.getLocalIdentity().getIDURL(as_original=True)
        if _Debug:
            lg.args(_DebugLevel, my_idurl=my_idurl)

        def _verify(xmlsrc=None):
            if not xmlsrc:
                lg.err('my current identity server not healthy')
                self.NeedPropagate = True
                self.automat('check-synchronize')
                return
            remote_ident = identity.identity(xmlsrc=xmlsrc)
            if not remote_ident.isCorrect() or not remote_ident.Valid():
                lg.warn('my current identity server responded with bad identity file')
                self.NeedPropagate = True
                self.automat('check-synchronize')
                return
            if remote_ident.getIDURL(as_original=True) != my_idurl:
                lg.warn('my current identity server responded with unknown identity')
                self.NeedPropagate = True
                self.automat('check-synchronize')
                return
            if _Debug:
                lg.dbg(_DebugLevel, 'my current identity server is healthy')

        last_time = identitycache.last_time_cached(my_idurl)
        if last_time and time.time() - last_time < config.conf().getInt('services/identity-propagate/health-check-interval-seconds'):
            if _Debug:
                lg.dbg(_DebugLevel, 'skip health check of my current identity server, last time cached %f seconds ago' % (time.time() - last_time))
            return

        d = identitycache.immediatelyCaching(my_idurl, try_other_sources=False)
        d.addCallback(_verify)
        d.addErrback(lambda _: _verify() and None)



#!/usr/bin/python
# local_service.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (local_service.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: local_service.

.. role:: red

BitDust local_service() Automat

.. raw:: html

    <a href="local_service.png" target="_blank">
    <img src="local_service.png" style="max-width:100%;">
    </a>

EVENTS:
    * :red:`depend-service-stopped`
    * :red:`service-depend-off`
    * :red:`service-failed`
    * :red:`service-not-installed`
    * :red:`service-started`
    * :red:`service-stopped`
    * :red:`shutdown`
    * :red:`start`
    * :red:`stop`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 12

#------------------------------------------------------------------------------

import os

#------------------------------------------------------------------------------

from twisted.internet.defer import Deferred  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.services import driver

#------------------------------------------------------------------------------


class LocalService(automat.Automat):
    """
    This class implements all the functionality of the ``local_service()``
    state machine.
    """

    fast = False
    suspended = False

    service_name = ''
    config_path = ''
    data_dir_required = False
    stop_when_failed = False
    start_suspended = False

    def __init__(self):
        if not self.service_name:
            raise driver.RequireSubclass()
        if self.service_name in list(driver.services().keys()):
            raise driver.ServiceAlreadyExist(self.service_name)
        self.result_deferred = None
        if self.data_dir_required:
            my_data_dir_path = self.data_dir_path()
            if not os.path.isdir(my_data_dir_path):
                os.makedirs(my_data_dir_path)
        automat.Automat.__init__(
            self,
            name=self.service_name,
            state='OFF',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
        )

    def to_json(self):
        j = super().to_json()
        j.update({
            'name': self.service_name,
            'enabled': self.enabled(),
            'installed': self.installed(),
            'config_path': self.config_path,
            'depends': self.dependent_on(),
        })
        return j

    #------------------------------------------------------------------------------

    def dependent_on(self):
        return []

    def installed(self):
        return True

    def enabled(self):
        from bitdust.main import config
        return config.conf().getBool(self.config_path)

    def start(self):
        raise driver.RequireSubclass()

    def stop(self):
        raise driver.RequireSubclass()

    def suspend(self, *args, **kwargs):
        if self.suspended:
            raise driver.ServiceAlreadySuspended()
        ret = self.on_suspend(*args, **kwargs)
        if ret is not None:
            self.suspended = True
        return ret

    def resume(self, *args, **kwargs):
        if not self.suspended:
            raise driver.ServiceWasNotSuspended()
        ret = self.on_resume(*args, **kwargs)
        if ret is not None:
            self.suspended = False
        return ret

    def request(self, json_payload, newpacket, info):
        raise driver.RequireSubclass()

    def cancel(self, json_payload, newpacket, info):
        raise driver.RequireSubclass()

    def on_suspend(self, *args, **kwargs):
        return True

    def on_resume(self, *args, **kwargs):
        return True

    def health_check(self):
        return True

    def network_configuration(self):
        return None

    def attached_dht_layers(self):
        return []

    def add_callback(self, cb):
        if not self.result_deferred:
            self.result_deferred = Deferred()
        if isinstance(cb, Deferred):
            self.result_deferred.addCallback(cb.callback)
        else:
            self.result_deferred.addCallback(lambda *a, **kw: cb(*a, **kw))
        return self.result_deferred

    def data_dir_path(self):
        from bitdust.main import settings
        return settings.ServiceDir(self.service_name)

    #------------------------------------------------------------------------------

    def A(self, event, *args, **kwargs):
        #---ON---
        if self.state == 'ON':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doStopService(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'stop':
                self.state = 'INFLUENCE'
                self.doSetCallback(*args, **kwargs)
                self.doStopDependentServices(*args, **kwargs)
        #---OFF---
        elif self.state == 'OFF':
            if event == 'stop':
                self.doSetCallback(*args, **kwargs)
                self.doNotifyStopped(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'start':
                self.state = 'STARTING'
                self.NeedStart = False
                self.NeedStop = False
                self.doSetCallback(*args, **kwargs)
                self.doStartService(*args, **kwargs)
        #---NOT_INSTALLED---
        elif self.state == 'NOT_INSTALLED':
            if event == 'stop':
                self.doSetCallback(*args, **kwargs)
                self.doNotifyNotInstalled(*args, **kwargs)
            elif event == 'start':
                self.state = 'STARTING'
                self.doSetCallback(*args, **kwargs)
                self.doStartService(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #---INFLUENCE---
        elif self.state == 'INFLUENCE':
            if event == 'start':
                self.doSetCallback(*args, **kwargs)
                self.NeedStart = True
            elif event == 'stop':
                self.doSetCallback(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doStopService(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'depend-service-stopped' and self.isAllDependsStopped(*args, **kwargs) and self.NeedStart:
                self.state = 'STARTING'
                self.NeedStart = False
                self.doStopService(*args, **kwargs)
                self.doNotifyStopped(*args, **kwargs)
                self.doStartService(*args, **kwargs)
            elif event == 'depend-service-stopped' and self.isAllDependsStopped(*args, **kwargs) and not self.NeedStart:
                self.state = 'STOPPING'
                self.doStopService(*args, **kwargs)
        #---STARTING---
        elif self.state == 'STARTING':
            if event == 'stop':
                self.doSetCallback(*args, **kwargs)
                self.NeedStop = True
            elif event == 'start':
                self.doSetCallback(*args, **kwargs)
            elif event == 'service-started' and self.NeedStop:
                self.state = 'INFLUENCE'
                self.NeedStop = False
                self.doStopDependentServices(*args, **kwargs)
            elif event == 'service-failed':
                self.state = 'OFF'
                self.doNotifyFailed(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doStopService(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'service-not-installed':
                self.state = 'NOT_INSTALLED'
                self.doNotifyNotInstalled(*args, **kwargs)
            elif event == 'service-depend-off':
                self.state = 'DEPENDS_OFF'
                self.doNotifyDependsOff(*args, **kwargs)
            elif event == 'service-started' and not self.NeedStop:
                self.state = 'ON'
                self.doNotifyStarted(*args, **kwargs)
        #---DEPENDS_OFF---
        elif self.state == 'DEPENDS_OFF':
            if event == 'stop':
                self.doSetCallback(*args, **kwargs)
                self.doNotifyDependsOff(*args, **kwargs)
            elif event == 'start':
                self.state = 'STARTING'
                self.doSetCallback(*args, **kwargs)
                self.doStartService(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #---STOPPING---
        elif self.state == 'STOPPING':
            if event == 'start':
                self.doSetCallback(*args, **kwargs)
                self.NeedStart = True
            elif event == 'service-stopped':
                self.state = 'OFF'
                self.doNotifyStopped(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        return None

    def isAllDependsStopped(self, *args, **kwargs):
        """
        Condition method.
        """
        for svc in driver.services().values():
            if self.service_name in svc.dependent_on():
                if svc.state != 'OFF' and svc.state != 'DEPENDS_OFF' and svc.state != 'NOT_INSTALLED':
                    if _Debug:
                        lg.out(_DebugLevel, '    dependent %r not stopped yet, %r will have to wait' % (svc, self))
                    return False
        return True

    def doStartService(self, *args, **kwargs):
        """
        Action method.
        """
        if not self.installed():
            self.automat('service-not-installed')
            return
        depends_results = []
        for depend_name in self.dependent_on():
            depend_service = driver.services().get(depend_name, None)
            if depend_service is None:
                depends_results.append((depend_name, 'not found'))
                continue
            if depend_service.state != 'ON':
                depends_results.append((depend_name, 'not started'))
                continue
        if len(depends_results) > 0:
            self.automat('service-depend-off', depends_results)
            return
        if _Debug:
            lg.out(_DebugLevel, '[%s] STARTING' % self.service_name)
        self.suspended = bool(self.start_suspended)
        try:
            result = self._do_start()
        except Exception as exc:
            lg.exc()
            self.automat('service-failed', exc)
            return
        if isinstance(result, Deferred):
            result.addCallback(lambda x: self.automat('service-started'))
            result.addErrback(lambda err: self.automat('service-failed', err))
            return
        if result:
            self.automat('service-started')
        else:
            lg.warn('failed to start %r, result from .start() method is %r' % (self, result))
            self.automat('service-failed', Exception('service %r failed to start' % self))

    def doStopService(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, '[%s] STOPPING' % self.service_name)
        self.suspended = False
        try:
            result = self._do_stop()
        except:
            lg.exc()
            self.automat('service-stopped', 'exception during stopping [%s]' % self.service_name)
            return
        if isinstance(result, Deferred):
            result.addBoth(lambda x: self.automat('service-stopped', x))
        else:
            self.automat('service-stopped', result)

    def doSetCallback(self, *args, **kwargs):
        """
        Action method.
        """
        if args and args[0]:
            self.add_callback(args[0])

    def doStopDependentServices(self, *args, **kwargs):
        """
        Action method.
        """
        count = 0
        for svc in driver.services().values():
            if self.service_name in svc.dependent_on():
                if _Debug:
                    lg.out(_DebugLevel, '%r sends "stop" to %r' % (self, svc))
                if svc.installed():
                    svc.automat('stop')
                    count += 1
        if count == 0:
            self.automat('depend-service-stopped')

    def doNotifyStarted(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.args(_DebugLevel, service=self, result='started')
        if self.result_deferred:
            self.result_deferred.callback('started')
            self.result_deferred = None
        driver.on_service_callback('started', self.service_name)

    def doNotifyStopped(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.args(_DebugLevel, service=self, result='stopped')
        if self.result_deferred:
            self.result_deferred.callback('stopped')
            self.result_deferred = None
        driver.on_service_callback('stopped', self.service_name)

    def doNotifyNotInstalled(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.args(_DebugLevel, service=self, result='not_installed')
        if self.result_deferred:
            self.result_deferred.callback('not_installed')
            self.result_deferred = None
        driver.on_service_callback('not_installed', self.service_name)

    def doNotifyFailed(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.args(_DebugLevel, service=self, result='failed')
        if self.stop_when_failed:
            try:
                self.stop()
            except:
                lg.exc()
        if self.result_deferred:
            self.result_deferred.callback('failed')
            self.result_deferred = None
        driver.on_service_callback('failed', self.service_name)

    def doNotifyDependsOff(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.args(_DebugLevel, service=self, result='depends_off')
        if self.result_deferred:
            self.result_deferred.callback('depends_off')
            self.result_deferred = None
        driver.on_service_callback('depends_off', self.service_name)

    def doDestroyMe(self, *args, **kwargs):
        """
        Action method.
        """
        self.result_deferred = None
        self.destroy()

    def _do_start(self):
        return self.start()

    def _do_stop(self):
        return self.stop()


#------------------------------------------------------------------------------


class SlowStartingLocalService(LocalService):
    def _do_start(self, **kwargs):
        if _Debug:
            lg.args(_DebugLevel, service_name=self.service_name)
        if getattr(self, 'starting_deferred', None):
            raise Exception('service already starting')
        self.starting_deferred = Deferred()
        self.starting_deferred.addErrback(lambda err: lg.warn('service %r was not started: %r' % (self.service_name, err.getErrorMessage() if err else 'unknown reason')))
        return self.start()

    def _do_stop(self):
        if _Debug:
            lg.args(_DebugLevel, service_name=self.service_name)
        if self.starting_deferred:
            if not self.starting_deferred.called:
                self.starting_deferred.errback(Exception('service was stopped before starting process finished'))
            self.starting_deferred = None
        return self.stop()

    def confirm_service_started(self, result=True):
        if _Debug:
            lg.args(_DebugLevel, service_name=self.service_name, result=result, starting_deferred=self.starting_deferred)
        if self.starting_deferred:
            if not self.starting_deferred.called:
                if result is True:
                    self.starting_deferred.callback(True)
                else:
                    if isinstance(result, Exception):
                        self.starting_deferred.errback(result)
                    else:
                        self.starting_deferred.callback(result)
            self.starting_deferred = None



#!/usr/bin/env python
# msgformat.py
#
# Copyright (C) 2007-2008 Francois Aucamp, Meraka Institute, CSIR
# See AUTHORS for all authors and contact information.
#
# License: GNU Lesser General Public License, version 3 or later; see COPYING
#          included in this archive for details.
#
# This library is free software, distributed under the terms of
# the GNU Lesser General Public License Version 3, or any later version.
# See the COPYING file included in this archive
#
# The docstrings in this module contain epytext markup; API documentation
# may be created by processing this file with epydoc: http://epydoc.sf.net

from __future__ import absolute_import
from six.moves import range

from . import msgtypes  # @UnresolvedImport


class MessageTranslator(object):
    """
    Interface for RPC message translators/formatters.

    Classes inheriting from this should provide a translation services
    between the classes used internally by this Kademlia implementation
    and the actual data that is transmitted between nodes.
    """
    def fromPrimitive(self, msgPrimitive):
        """
        Create an RPC Message from a message's string representation.

        @param msgPrimitive: The unencoded primitive representation of a message
        @type msgPrimitive: str, int, list or dict

        @return: The translated message object
        @rtype: entangled.kademlia.msgtypes.Message
        """

    def toPrimitive(self, message):
        """
        Create a string representation of a message.

        @param message: The message object
        @type message: msgtypes.Message

        @return: The message's primitive representation in a particular
                 messaging format
        @rtype: str, int, list or dict
        """


class DefaultFormat(MessageTranslator):
    """
    The default on-the-wire message format for this library.
    """
    typeRequest, typeResponse, typeError = list(range(3))
    headerType, headerMsgID, headerNodeID, headerPayload, headerArgs = list(range(5))

    def fromPrimitive(self, msgPrimitive):
        msgType = msgPrimitive[self.headerType]
        if msgType == self.typeRequest:
            msg = msgtypes.RequestMessage(msgPrimitive[self.headerNodeID], msgPrimitive[self.headerPayload], msgPrimitive[self.headerArgs], msgPrimitive[self.headerMsgID])
        elif msgType == self.typeResponse:
            msg = msgtypes.ResponseMessage(msgPrimitive[self.headerMsgID], msgPrimitive[self.headerNodeID], msgPrimitive[self.headerPayload])
        elif msgType == self.typeError:
            msg = msgtypes.ErrorMessage(msgPrimitive[self.headerMsgID], msgPrimitive[self.headerNodeID], msgPrimitive[self.headerPayload], msgPrimitive[self.headerArgs])
        else:
            # Unknown message, no payload
            msg = msgtypes.Message(msgPrimitive[self.headerMsgID], msgPrimitive[self.headerNodeID])
        return msg

    def toPrimitive(self, message):
        msg = {self.headerMsgID: message.id, self.headerNodeID: message.nodeID}
        if isinstance(message, msgtypes.RequestMessage):
            msg[self.headerType] = self.typeRequest
            msg[self.headerPayload] = message.request
            msg[self.headerArgs] = message.args
        elif isinstance(message, msgtypes.ErrorMessage):
            msg[self.headerType] = self.typeError
            msg[self.headerPayload] = message.exceptionType
            msg[self.headerArgs] = message.response
        elif isinstance(message, msgtypes.ResponseMessage):
            msg[self.headerType] = self.typeResponse
            msg[self.headerPayload] = message.response
        return msg


class MultiLayerFormat(MessageTranslator):
    typeRequest, typeResponse, typeError, typeQuestion = list(range(4))
    headerType, headerMsgID, headerNodeID, headerPayload, headerArgs, headerLayer = list(range(6))

    def fromPrimitive(self, msgPrimitive):
        msgType = msgPrimitive[self.headerType]
        if msgType == self.typeRequest:
            layerID = msgPrimitive[self.headerLayer] if self.headerLayer in msgPrimitive else 0
            msg = msgtypes.RequestMessage(msgPrimitive[self.headerNodeID], msgPrimitive[self.headerPayload], msgPrimitive[self.headerArgs], msgPrimitive[self.headerMsgID], layerID=layerID)
        elif msgType == self.typeResponse:
            layerID = msgPrimitive[self.headerLayer] if self.headerLayer in msgPrimitive else 0
            msg = msgtypes.ResponseMessage(msgPrimitive[self.headerMsgID], msgPrimitive[self.headerNodeID], msgPrimitive[self.headerPayload], layerID=layerID)
        elif msgType == self.typeError:
            layerID = msgPrimitive[self.headerLayer] if self.headerLayer in msgPrimitive else 0
            msg = msgtypes.ErrorMessage(msgPrimitive[self.headerMsgID], msgPrimitive[self.headerNodeID], msgPrimitive[self.headerPayload], msgPrimitive[self.headerArgs], layerID=layerID)
        elif msgType == self.typeQuestion:
            layerID = msgPrimitive[self.headerLayer] if self.headerLayer in msgPrimitive else 0
            msg = msgtypes.QuestionMessage(msgPrimitive[self.headerNodeID], msgPrimitive[self.headerPayload], msgPrimitive[self.headerArgs], msgPrimitive[self.headerMsgID], layerID=layerID)
        else:
            # Unknown message, no payload
            msg = msgtypes.Message(msgPrimitive[self.headerMsgID], msgPrimitive[self.headerNodeID])
        return msg

    def toPrimitive(self, message):
        msg = {self.headerMsgID: message.id, self.headerNodeID: message.nodeID}
        if isinstance(message, msgtypes.RequestMessage):
            msg[self.headerType] = self.typeRequest
            msg[self.headerPayload] = message.request
            msg[self.headerArgs] = message.args
            msg[self.headerLayer] = message.layerID
        elif isinstance(message, msgtypes.ErrorMessage):
            msg[self.headerType] = self.typeError
            msg[self.headerPayload] = message.exceptionType
            msg[self.headerArgs] = message.response
            msg[self.headerLayer] = message.layerID
        elif isinstance(message, msgtypes.ResponseMessage):
            msg[self.headerType] = self.typeResponse
            msg[self.headerPayload] = message.response
            msg[self.headerLayer] = message.layerID
        elif isinstance(message, msgtypes.QuestionMessage):
            msg[self.headerType] = self.typeQuestion
            msg[self.headerPayload] = message.request
            msg[self.headerArgs] = message.args
            msg[self.headerLayer] = message.layerID
        return msg



#!/usr/bin/python
# propagate.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (propagate.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: propagate.

.. role:: red

When a user starts up he needs to run the stun.py to check what his IP is,
and if it has changed he needs to generate a new identity and send it to
his identity server and all of his contacts.

We also just request new copies of all identities from their servers when
we start up. This is simple and effective.

We should try contacting each contact every hour and if we have not been
able to contact them in 2 or 3 hours then fetch copy of identity from
their server.
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import sys

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in propagate.py')

from twisted.internet.defer import DeferredList, Deferred

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import nameurl
from bitdust.lib import net_misc
from bitdust.lib import strng
from bitdust.lib import packetid

from bitdust.contacts import contactsdb
from bitdust.contacts import identitycache

from bitdust.main import settings

from bitdust.userid import my_id
from bitdust.userid import id_url

from bitdust.p2p import commands
from bitdust.p2p import p2p_stats

from bitdust.crypt import signed

from bitdust.transport import gateway
from bitdust.transport import packet_out

from bitdust.dht import dht_records

#------------------------------------------------------------------------------

_SlowSendIsWorking = False
_PropagateCounter = 0
_StartupPropagateList = set()

#------------------------------------------------------------------------------


def init():
    """
    Need to call that at start up to link with transport_control.
    """
    if _Debug:
        lg.out(_DebugLevel, 'propagate.init')
    # callback.add_finish_file_sending_callback(OnFileSent)


def shutdown():
    if _Debug:
        lg.out(_DebugLevel, 'propagate.shutdown')


def startup_list():
    global _StartupPropagateList
    return _StartupPropagateList


#------------------------------------------------------------------------------


def propagate(selected_contacts, ack_handler=None, wide=False, refresh_cache=False, wait_packets=False, response_timeout=None):
    """
    Run the "propagate" process.

    First need to fetch ``selected_contacts`` IDs from id servers. And
    then send our Identity file to that contacts.
    """
    if response_timeout is None:
        response_timeout = settings.P2PTimeOut()
    if _Debug:
        lg.out(_DebugLevel, 'propagate.propagate to %d contacts' % len(selected_contacts))
    result = Deferred()

    def contacts_fetched(x):
        res = SendToIDs(
            idlist=selected_contacts,
            ack_handler=ack_handler,
            wide=wide,
            wait_packets=wait_packets,
            response_timeout=response_timeout,
        )
        if _Debug:
            lg.out(_DebugLevel, 'propagate.contacts_fetched with %d identities, sending my identity to %d remote nodes: %r' % (len(x), len(selected_contacts), res))
        if wait_packets:
            if not res:
                result.callback([])
                return result
            res.addCallback(result.callback)
            res.addErrback(result.errback)
            return result
        result.callback(list(selected_contacts))
        return result

    fetch(list_ids=selected_contacts, refresh_cache=refresh_cache).addBoth(contacts_fetched)
    return result


def fetch(list_ids, refresh_cache=False, timeout=15, try_other_sources=True):
    """
    Request a list of identity files.
    """
    if _Debug:
        lg.out(_DebugLevel, 'propagate.fetch %d identities' % len(list_ids))
    dl = []
    for url in list_ids:
        if not url:
            continue
        if identitycache.FromCache(url) and not refresh_cache:
            continue
        dl.append(identitycache.immediatelyCaching(
            idurl=id_url.to_original(url),
            timeout=timeout,
            try_other_sources=try_other_sources,
        ))
    return DeferredList(dl, consumeErrors=True)


def start(ack_handler=None, wide=False, refresh_cache=False, include_all=True, include_enabled=True, include_startup=False, wait_packets=False, response_timeout=None):
    """
    Call ``propagate()`` for all known contacts or only for those which are related to enabled/active services.
    """
    if response_timeout is None:
        response_timeout = settings.P2PTimeOut()
    selected_contacts = set(filter(None, contactsdb.contacts_remote(include_all=include_all, include_enabled=include_enabled)))
    if include_startup and startup_list():
        lg.warn('going to propagate my identity also to %d nodes from startup list' % len(startup_list()))
        selected_contacts.update(startup_list())
        startup_list().clear()
    if _Debug:
        lg.args(_DebugLevel, wide=wide, refresh_cache=refresh_cache, all=include_all, enabled=include_enabled, selected=len(selected_contacts))
    return propagate(
        selected_contacts=list(selected_contacts),
        ack_handler=ack_handler,
        wide=wide,
        refresh_cache=refresh_cache,
        wait_packets=wait_packets,
        response_timeout=response_timeout,
    )


def suppliers(ack_handler=None, wide=False, customer_idurl=None, wait_packets=False):
    """
    Call ``propagate()`` for all suppliers.
    """
    if _Debug:
        lg.out(_DebugLevel, 'propagate.suppliers')
    return propagate(
        selected_contacts=contactsdb.suppliers(customer_idurl=customer_idurl),
        ack_handler=ack_handler,
        wide=wide,
        wait_packets=wait_packets,
    )


def customers(ack_handler=None, wide=False, wait_packets=False):
    """
    Call ``propagate()`` for all known customers.
    """
    if _Debug:
        lg.out(_DebugLevel, 'propagate.customers')
    return propagate(
        selected_contacts=contactsdb.customers(),
        ack_handler=ack_handler,
        wide=wide,
        wait_packets=wait_packets,
    )


def single(idurl, ack_handler=None, wide=False, fail_handler=None):
    """
    Do "propagate" for a single contact.
    """
    d = FetchSingle(idurl)
    d.addCallback(lambda x: SendToIDs(
        [
            idurl,
        ],
        ack_handler=ack_handler,
        wide=wide,
    ))
    if ack_handler:
        d.addErrback(fail_handler or ack_handler)
    return d


def update():
    """
    A wrapper of ``SendServers()`` method.
    """
    if _Debug:
        lg.out(_DebugLevel, 'propagate.update')
    return SendServers()


def write_to_dht():
    """

    """
    if _Debug:
        lg.out(_DebugLevel, 'propagate.write_to_dht')
    LocalIdentity = my_id.getLocalIdentity()
    return dht_records.set_identity(
        LocalIdentity.getIDURL(),
        LocalIdentity.serialize(as_text=True),
    )


#------------------------------------------------------------------------------


def FetchSingle(idurl):
    """
    Fetch single identity file from given ``idurl``.
    """
    if _Debug:
        lg.out(_DebugLevel, 'propagate.fetch_single ' + idurl)
    return identitycache.scheduleForCaching(idurl)


def Fetch(idslist):
    """
    Just a wrapper for ``fetch()`` method.
    """
    return fetch(idslist)


def FetchSuppliers(customer_idurl=None):
    """
    Fetch identity files of all supplier.
    """
    return fetch(contactsdb.suppliers(customer_idurl=customer_idurl))


def FetchCustomers():
    """
    Fetch identity files of all customers.
    """
    return fetch(contactsdb.customers())


#------------------------------------------------------------------------------


def SendServers():
    """
    My identity file can be stored in different locations, see the "sources" field.
    So I can use different identity servers to store more secure and reliable. This
    method will send my identity file to all my identity servers with HTTP POST method.
    """
    LocalIdentity = my_id.getLocalIdentity()
    payload = LocalIdentity.serialize(as_text=False)
    dlist = []
    for idurl in LocalIdentity.getSources(as_originals=True):
        _, host, webport, filename = nameurl.UrlParse(idurl)
        url = net_misc.pack_address(
            (
                host,
                webport,
            ),
            proto='http',
        )
        dlist.append(net_misc.http_post_data(
            url=url,
            data=payload,
            connectTimeout=15,
        ))
        if _Debug:
            lg.args(_DebugLevel, url=url, filename=filename, size=len(payload))
    dl = DeferredList(dlist, consumeErrors=True)
    return dl


def SendSuppliers(customer_idurl=None):
    """
    Send my identity file to all my suppliers, calls to ``SendToIDs()`` method.
    """
    if _Debug:
        lg.out(_DebugLevel, 'propagate.SendSuppliers')
    SendToIDs(contactsdb.suppliers(customer_idurl=customer_idurl), ack_handler=HandleSuppliersAck, wide=True)


def SendCustomers():
    """
    Calls ``SendToIDs()`` to send identity to all my customers.
    """
    if _Debug:
        lg.out(_DebugLevel, 'propagate.SendCustomers')
    SendToIDs(contactsdb.customers(), ack_handler=HandleCustomersAck, wide=True)


def SlowSendSuppliers(delay=1, customer_idurl=None):
    """
    Doing same thing, but puts delays before sending to every next supplier.

    This is used when need to "ping" suppliers.
    """
    global _SlowSendIsWorking
    if _SlowSendIsWorking:
        if _Debug:
            lg.out(_DebugLevel, 'propagate.SlowSendSuppliers  is working at the moment. skip.')
        return
    if _Debug:
        lg.out(_DebugLevel, 'propagate.SlowSendSuppliers delay=%s' % str(delay))

    def _send(index, payload, delay):
        global _SlowSendIsWorking
        idurl = contactsdb.supplier(index, customer_idurl=customer_idurl)
        if not idurl:
            _SlowSendIsWorking = False
            return
        # transport_control.ClearAliveTime(idurl)
        SendToID(idurl, Payload=payload, wide=True)
        reactor.callLater(delay, _send, index + 1, payload, delay)  # @UndefinedVariable

    _SlowSendIsWorking = True
    payload = strng.to_bin(my_id.getLocalIdentity().serialize())
    _send(0, payload, delay)


def SlowSendCustomers(delay=1):
    """
    Same, "slowly" send my identity file to all my customers.
    """

    global _SlowSendIsWorking
    if _SlowSendIsWorking:
        if _Debug:
            lg.out(_DebugLevel, 'propagate.SlowSendCustomers  slow send is working at the moment. skip.')
        return
    if _Debug:
        lg.out(_DebugLevel, 'propagate.SlowSendCustomers delay=%s' % str(delay))

    def _send(index, payload, delay):
        global _SlowSendIsWorking
        idurl = contactsdb.customer(index)
        if not idurl:
            _SlowSendIsWorking = False
            return
        # transport_control.ClearAliveTime(idurl)
        SendToID(idurl, Payload=payload, wide=True)
        reactor.callLater(delay, _send, index + 1, payload, delay)  # @UndefinedVariable

    _SlowSendIsWorking = True
    payload = strng.to_bin(my_id.getLocalIdentity().serialize())
    _send(0, payload, delay)


def HandleSuppliersAck(ackpacket, info):
    """
    Called when supplier is "Acked" to my after call to ``SendSuppliers()``.
    """
    if _Debug:
        lg.out(_DebugLevel, 'propagate.HandleSupplierAck %s' % ackpacket.OwnerID)


def HandleCustomersAck(ackpacket, info):
    """
    Called when supplier is "Acked" to my after call to ``SendCustomers()``.
    """
    if _Debug:
        lg.out(_DebugLevel, 'propagate.HandleCustomerAck %s' % ackpacket.OwnerID)


def HandleAck(ackpacket, info):
    if _Debug:
        lg.out(_DebugLevel, 'propagate.HandleAck %r %r' % (ackpacket, info))


def HandleTimeOut(pkt_out):
    if _Debug:
        lg.out(_DebugLevel, 'propagate.HandleTimeOut %r' % pkt_out)


def OnFileSent(pkt_out, item, status, size, error_message):
    return False


def SendToID(
    idurl,
    Payload=None,
    wide=False,
    ack_handler=None,
    timeout_handler=None,
    response_timeout=None,
):
    """
    Create ``packet`` with my Identity file and calls
    ``transport.gateway.outbox()`` to send it.
    """
    global _PropagateCounter
    if _Debug:
        lg.out(_DebugLevel, 'propagate.SendToID [%s] wide=%s' % (nameurl.GetName(idurl), str(wide)))
    if response_timeout is None:
        response_timeout = settings.P2PTimeOut()
    if ack_handler is None:
        ack_handler = HandleAck
    if timeout_handler is None:
        timeout_handler = HandleTimeOut
    thePayload = Payload
    if thePayload is None:
        thePayload = strng.to_bin(my_id.getLocalIdentity().serialize())
    p = signed.Packet(
        Command=commands.Identity(),
        OwnerID=my_id.getIDURL(),
        CreatorID=my_id.getIDURL(),
        PacketID=('propagate:%d:%s' % (_PropagateCounter, packetid.UniqueID())),
        Payload=thePayload,
        RemoteID=idurl,
    )
    _PropagateCounter += 1
    result = gateway.outbox(
        p,
        wide,
        response_timeout=response_timeout,
        callbacks={
            commands.Ack(): ack_handler,
            commands.Fail(): ack_handler,
            None: timeout_handler,
        },
    )
    if wide:
        # this is a ping packet - need to clear old info
        p2p_stats.ErasePeerProtosStates(idurl)
        p2p_stats.EraseMyProtosStates(idurl)
    return result


def SendToIDs(idlist, wide=False, ack_handler=None, timeout_handler=None, response_timeout=None, wait_packets=False):
    """
    Same, but send to many IDs and also check previous packets to not re-send.
    """
    global _PropagateCounter
    if response_timeout is None:
        response_timeout = settings.P2PTimeOut()
    if ack_handler is None:
        ack_handler = HandleAck
    if timeout_handler is None:
        timeout_handler = HandleTimeOut
    LocalIdentity = my_id.getLocalIdentity()
    Payload = strng.to_bin(LocalIdentity.serialize())
    if _Debug:
        lg.out(_DebugLevel, 'propagate.SendToIDs to %d users, rev=%r wide=%s' % (len(idlist), LocalIdentity.getRevisionValue(), wide))
    alreadysent = set()
    totalsent = 0
    inqueue = {}
    found_previous_packets = 0
    for pkt_out in packet_out.queue():
        if id_url.is_in(pkt_out.remote_idurl, idlist, as_field=False):
            if pkt_out.description.count('Identity'):
                if pkt_out.remote_idurl not in inqueue:
                    inqueue[pkt_out.remote_idurl] = 0
                inqueue[pkt_out.remote_idurl] += 1
                found_previous_packets += 1
    wait_list = []
    for contact in idlist:
        if not contact:
            continue
        if contact in alreadysent:
            # just want to send once even if both customer and supplier
            continue
        if contact in inqueue and inqueue[contact] > 2:
            # now only 2 protocols is working: tcp and udp
            if _Debug:
                lg.out(_DebugLevel, '        skip sending [Identity] to %s, packet already in the queue' % contact)
            continue
        p = signed.Packet(
            Command=commands.Identity(),
            OwnerID=my_id.getIDURL(),
            CreatorID=my_id.getIDURL(),
            PacketID=('propagate:%d:%s' % (_PropagateCounter, packetid.UniqueID())),
            Payload=Payload,
            RemoteID=contact,
        )
        _PropagateCounter += 1
        if _Debug:
            lg.out(_DebugLevel, '        sending %r to %s' % (p, nameurl.GetName(contact)))
        res = gateway.outbox(
            p,
            wide,
            response_timeout=response_timeout,
            callbacks={
                commands.Ack(): ack_handler,
                commands.Fail(): ack_handler,
                None: timeout_handler,
            },
        )
        if not res:
            lg.warn('my Identity() was not sent to %r' % contact)
            continue
        if wide:
            # this is a ping packet - need to clear old info
            p2p_stats.ErasePeerProtosStates(contact)
            p2p_stats.EraseMyProtosStates(contact)
        alreadysent.add(contact)
        totalsent += 1
        if wait_packets and res:
            if isinstance(res, Deferred):
                wait_list.append(res)
            elif res.finished_deferred and isinstance(res.finished_deferred, Deferred):
                wait_list.append(res.finished_deferred)
    del alreadysent
    if not wait_packets:
        return totalsent
    return DeferredList(wait_list, consumeErrors=True)


#------------------------------------------------------------------------------


def ping_suppliers(customer_idurl=None, timeout=None):
    if timeout is None:
        timeout = settings.P2PTimeOut()
    from bitdust.p2p import online_status
    l = []
    for supplier_idurl in contactsdb.suppliers(customer_idurl=customer_idurl):
        if supplier_idurl:
            l.append(online_status.ping(idurl=supplier_idurl, ack_timeout=timeout, channel='ping_suppliers', keep_alive=True))
    return DeferredList(l, consumeErrors=True)


def ping_customers(timeout=None):
    if timeout is None:
        timeout = settings.P2PTimeOut()
    from bitdust.p2p import online_status
    l = []
    for customer_idurl in contactsdb.customers():
        if customer_idurl:
            l.append(online_status.ping(idurl=customer_idurl, ack_timeout=timeout, channel='ping_customers', keep_alive=True))
    return DeferredList(l, consumeErrors=True)


def ping_nodes(idurl_list, timeout=None, channel='ping_nodes', keep_alive=True):
    if timeout is None:
        timeout = settings.P2PTimeOut()
    from bitdust.p2p import online_status
    l = []
    for idurl in idurl_list:
        if idurl:
            l.append(online_status.ping(idurl=idurl, ack_timeout=timeout, channel=channel, keep_alive=keep_alive))
    return DeferredList(l, consumeErrors=True)
